"""
Provides a RETURNN wrapper around `warp-transducer`:
  https://github.com/1ytic/warp-rna

Importing this module immediately compiles the library and TF module.
"""

import os
import typing
import tensorflow as tf
from tensorflow.python.framework import ops
from returnn.tf.util.basic import OpCodeCompiler


warprna_dir = os.path.dirname(os.path.abspath(__file__))
submodule_dir = os.path.join(warprna_dir, "warp-rna")
_tf_mod = None  # type: typing.Optional[typing.Any]


def is_checked_out():
    """
    Checks if the git submodule is checkout out.

    :rtype: bool
    """
    return os.path.isfile("%s/core.cu" % submodule_dir)


def init_warprna(verbose=False):
    """
    Initializes and compiles the library. Caches the TF module.

    :param bool verbose:
    """
    global _tf_mod
    if _tf_mod:
        return
    assert is_checked_out(), "submodule not checked out? Run `git submodule update --init --recursive`"
    enable_gpu = OpCodeCompiler.cuda_available()
    enable_cpu = os.path.exists("%s/core_cpu.cpp" % submodule_dir)

    src_files = ["%s/tensorflow_binding/src/warp_rna_op.cc" % submodule_dir]
    if enable_gpu:
        src_files.append("%s/core.cu" % submodule_dir)
    if enable_cpu:
        src_files.append("%s/core_cpu.cpp" % submodule_dir)
    src_code = ""
    for fn in src_files:
        f_code = open(fn).read()
        src_code += "\n// ------------ %s : BEGIN { ------------\n" % os.path.basename(fn)
        # https://gcc.gnu.org/onlinedocs/cpp/Line-Control.html#Line-Control
        src_code += '#line 1 "%s"\n' % os.path.basename(fn)
        src_code += f_code
        src_code += "\n// ------------ %s : END } --------------\n\n" % os.path.basename(fn)

    compiler = OpCodeCompiler(
        base_name="warprna_kernels",
        code_version=1,
        code=src_code,
        include_paths=(
            submodule_dir,
            "%s/tensorflow_binding/src" % submodule_dir,
        ),
        c_macro_defines={
            "WARPRNA_ENABLE_CPU": 1 if enable_cpu else None,
            "WARPRNA_ENABLE_GPU": 1 if enable_gpu else None,
        },
        is_cpp=True,
        use_cuda_if_available=enable_gpu,
        verbose=verbose,
    )
    tf_mod = compiler.load_tf_module()
    assert hasattr(tf_mod, "WarpRNA"), "content of mod: %r" % (dir(tf_mod),)
    _tf_mod = tf_mod
    return tf_mod


def rna_loss(log_probs, labels, input_lengths, label_lengths, blank_label=0):
    """Computes the RNA loss between a sequence of activations and a
    ground truth labeling.
    Args:
        log_probs: A 4-D Tensor of floats.  The dimensions
                     should be (B, T, U, V), where B is the minibatch index,
                     T is the time index, U is the prediction network sequence
                     length, and V indexes over activations for each
                     symbol in the alphabet.
        labels: A 2-D Tensor of ints, shape (B,U-1) a padded label sequences to make sure
                     labels for the minibatch are same length.
        input_lengths: A 1-D Tensor of ints, shape (B,), the number of time steps
                       for each sequence in the minibatch.
        label_lengths: A 1-D Tensor of ints, shape (B,), the length of each label
                       for each example in the minibatch.
        blank_label: int, scalar, the label value/index that the RNA
                     calculation should use as the blank label
    Returns:
        1-D float Tensor, the cost of each example in the minibatch
        (as negative log probabilities).
    """
    init_warprna()
    loss, _ = _tf_mod.warp_rna(
        log_probs, labels, input_lengths, label_lengths, tf.reduce_min(label_lengths), blank_label
    )
    return loss


@ops.RegisterGradient("WarpRNA")
def _warprna_loss_grad(op, grad_loss, _):
    grad = op.outputs[1]
    # NOTE since here we are batch first, cannot use _BroadcastMul
    grad_loss = tf.reshape(grad_loss, (-1, 1, 1, 1))
    return [grad_loss * grad, None, None, None, None]



"""
Array (Tensor) functions
"""

from __future__ import annotations
from typing import Optional, Union, Type, TypeVar, Sequence, Tuple
import logging
import numpy
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from ._backend import Backend, global_backend, get_backend_by_raw_tensor_type
from .types import RawTensorTypes
from . import _utils

T = TypeVar("T")

__all__ = [
    "convert_to_tensor",
    "constant",
    "copy",
    "cast",
    "merge_dims",
    "split_dims",
    "reshape",
    "split",
    "expand_dim",
    "expand_dims",
    "squeeze",
    "window",
    "concat",
    "concat_features",
    "pad",
    "cum_concat_step",
    "stack",
    "masked_select",
    "masked_scatter",
    "sequence_mask",
    "pack_padded",
    "pad_packed",
    "gather",
    "scatter",
    "scatter_argmax",
    "slice",
    "shift_right",
    "shift_left",
    "reverse_sequence",
    "where",
    "search_sorted",
    "sparse_to_dense",
    "one_hot",
]


def convert_to_tensor(
    value: Union[Tensor, T, RawTensorTypes, list, tuple],
    *,
    dims: Sequence[Dim] = None,
    dtype: Optional[str] = None,
    sparse_dim: Optional[Dim] = None,
    shape: Sequence[Dim] = None,
    device: Optional[str] = None,
    keep_scalar_on_cpu: bool = False,
    name: Optional[str] = None,
    _backend: Optional[Type[Backend]] = None,
) -> Tensor[T]:
    """
    :param value: tensor, or scalar raw tensor or some other scalar value
    :param dims:
    :param dtype:
    :param sparse_dim:
    :param shape: alias for dims, for some older code
    :param name:
    :param device:
    :param keep_scalar_on_cpu: if the value is already on the CPU, keep it there, even if `device` is sth else
    :param _backend:
    :return: tensor
    """
    if isinstance(value, Tensor):  # fast path
        return value
    if isinstance(value, (tuple, list)):
        value = numpy.array(value, dtype=dtype)
    if dims is None and shape is not None:
        dims = shape  # old code
    if isinstance(value, (int, float, complex, bool, str, numpy.number)):
        if _backend is None:
            _backend = global_backend
        if dims is None:
            dims = ()
        if dtype is None:
            if isinstance(value, bool):
                dtype = "bool"
            elif isinstance(value, int):
                dtype = rf.get_default_int_dtype()
            elif isinstance(value, float):
                dtype = rf.get_default_float_dtype()
            elif isinstance(value, str):
                dtype = "string"
            elif isinstance(value, numpy.number):
                dtype = value.dtype.name
            else:
                raise ValueError(f"number {value} type {type(value)} needs explicit `dtype` specification")
        if keep_scalar_on_cpu:
            device = "cpu"
    elif isinstance(value, numpy.ndarray):
        if _backend is None:
            # Small exception: Do not use the NumpyBackend but the global backend in this case.
            _backend = global_backend
        if dims is None:
            dims = [Dim(d) for d in value.shape]
        if dtype is None:
            dtype = value.dtype.name
    else:
        value_backend = get_backend_by_raw_tensor_type(type(value))
        if _backend is None:
            _backend = value_backend
        if dims is None:
            dims = [
                value_backend.get_new_dim_raw(value, d, name=(name or "const") + f"_dim{d}")
                for d in range(value_backend.get_ndim_raw(value))
            ]
        if dtype is None:
            dtype = value_backend.get_dtype_name_raw(value)
    return _backend.convert_to_tensor(
        value=value, dims=dims, dtype=dtype, sparse_dim=sparse_dim, device=device, name=name
    )


constant = convert_to_tensor  # alias for some older code


def copy(tensor: Tensor) -> Tensor:
    """
    :param tensor:
    :return: copy of tensor.
        In eager-based frameworks, it is really a copy.
        In graph-based frameworks, it might be just a copied reference if it would be immutable.
        This is really only relevant when operating on tensors which can conceptually be mutated,
        such as variables (:class:`Parameter`).
    """
    # noinspection PyProtectedMember
    return tensor._raw_backend.copy(tensor)


def cast(tensor: Tensor, dtype: str) -> Tensor:
    """
    :param tensor:
    :param dtype:
    :return: tensor with the same data, but with a different dtype
    """
    # noinspection PyProtectedMember
    return tensor._raw_backend.cast(tensor, dtype=dtype)


def merge_dims(
    source: Tensor,
    *,
    dims: Sequence[Dim],
    out_dim: Optional[Dim] = None,
) -> Tuple[Tensor, Dim]:
    """
    Merges a list of axes into a single one. (Flatten the dims.)
    E.g. input is (batch, width, height, dim) and dims=(width,height), then we get (batch, width*height, dim).
    Or input is (batch, time, height, dim) and axes=(height,dim), then we get (batch, time, height*dim).

    :func:`rf.split_dims` is the reverse operation.

    :param source:
    :param dims:
    :param out_dim:
    :return: tensor, out_dim
    """
    # noinspection PyProtectedMember
    return source._raw_backend.merge_dims(source, dims=dims, out_dim=out_dim)


def split_dims(
    source: Tensor,
    *,
    axis: Dim,
    dims: Sequence[Dim],
    pad_to_multiples: Optional[bool] = None,
    pad_value: Union[None, int, float] = None,
) -> Tensor:
    """
    Splits one axis into multiple axes.
    E.g. if you know that your feature-dim is composed by a window,
    i.e. the input is (batch, time, window * feature),
    you can set axis="F", dims=(window, -1),
    and you will get the output (batch, time, window, feature).

    If the split axis has a dynamic length,
    exactly one of the axes that we split into need to also have a dynamic length.
    You can e.g. use this to split the input dimension into smaller "chunks" of a fixed window size.
    E.g. you could have input (batch, time, feature) and set axis="T", dims=(-1, window),
    to get output (batch, split_time, window, feature).
    In this case, the exact sequence lengths are lost and everything is padded to multiples of the window size using
    the given padding value.
    Use :class:`ReinterpretDataLayer` to receive back the original sequence lengths after merging.

    Also see :func:`rf.merge_dims` which can undo this operation.

    :param source:
    :param axis: e.g. "F"
    :param dims: what the axis should be split into. e.g. (window, -1)
    :param pad_to_multiples: If true, input will be padded to the next multiple of the product of the
        static dims, such that splitting is actually possible.
        By default this is done iff the axis has a dynamic size
    :param pad_value: What pad value to use for pad_to_multiples
    :return: source with axis replaced by dims
    """
    # noinspection PyProtectedMember
    return source._raw_backend.split_dims(
        source, axis=axis, dims=dims, pad_to_multiples=pad_to_multiples, pad_value=pad_value
    )


def reshape(source: Tensor, in_dims: Sequence[Dim], out_dims: Sequence[Dim]) -> Tensor:
    """
    Wraps tf.reshape.

    You should use :func:`split_dims` or :func:`merge_dims`
    when you want to split or merge dimensions.
    This here is for doing any other kind of reshape.
    This can be used for clever indexing, slicing, padding tricks.

    :param source: e.g. (..., old_dims, ...)
    :param in_dims: the old dims which should be reshaped into new_dims.
      This should only cover those dims which should be reshaped,
      not all the dims of the source.
    :param out_dims: the new dims which should be reshaped from old_dims.
      This is excluding any of the other dims in the source.
    :return: e.g. (..., new_dims, ...)
    """
    # noinspection PyProtectedMember
    return source._raw_backend.reshape(source, in_dims=in_dims, out_dims=out_dims)


def split(source: Tensor, *, axis: Dim, out_dims: Sequence[Dim]) -> Tuple[Tensor, ...]:
    """
    Split the input on the specified axis (by default feature).
    Basically a wrapper around tf.split.

    :param source: {..., axis}
    :param axis: some static axis
    :param out_dims: list of dims where sum(out_dims) == axis
    :return: tuple of tensors, same amount as out_dims,
        with the same shape as source, but with the specified axis replaced by the out_dims
    """
    # noinspection PyProtectedMember
    return source._raw_backend.split(source, axis=axis, out_dims=out_dims)


def expand_dim(source: Tensor, dim: Dim) -> Tensor:
    """
    Expand the source by the given dimension.

    Note that this is *never* needed for broadcasting.
    All broadcasting should always happen automatically.

    This might be needed for convolution or concatenation.
    """
    # noinspection PyProtectedMember
    return source._raw_backend.expand_dim(source, dim=dim)


def expand_dims(source: Tensor, dims: Sequence[Dim]) -> Tensor:
    """
    Expand multiple dims, via :func:`expand_dim`.
    """
    for dim in dims:
        source = expand_dim(source, dim)
    return source


def squeeze(source: Tensor, axis: Dim) -> Tensor:
    """
    Removes the axis with dimension of extend 1 from the source.
    """
    assert axis.dimension == 1, f"squeeze {source}: axis {axis} is not of extend 1"
    # noinspection PyProtectedMember
    return source._raw_backend.squeeze(source, axis=axis)


def window(
    source: Tensor,
    *,
    spatial_dim: Dim,
    window_dim: Dim,
    window_right: Optional[Union[Dim, int]] = None,
    window_left: Optional[Union[Dim, int]] = None,
    padding: str = "same",
    pad_value: Optional[Union[int, float]] = None,
    stride: int = 1,
) -> Tuple[Tensor, Dim]:
    """
    Follows the same idea as RETURNN tf_util.windowed,
    using clever padding and reshaping.

    :param source:
    :param spatial_dim:
    :param window_dim:
    :param window_left:
    :param window_right:
    :param padding: "same" or "valid"
    :param pad_value:
    :param stride:
    :return: out, out_spatial_dim
    """
    assert window_dim.dimension is not None
    if padding == "same":
        out_spatial_dim = spatial_dim
        if window_right is not None:
            if isinstance(window_right, int):
                window_right = Dim(window_right, name="window_right")
            assert isinstance(window_right, Dim)
        if window_left is not None:
            if isinstance(window_left, int):
                window_left = Dim(window_left, name="window_left")
            assert isinstance(window_left, Dim)
        if window_right is None:
            if window_left is None:
                window_right = window_dim // 2
                window_left = window_dim.ceildiv_right(2) - 1
            else:
                window_right = window_dim - window_left - 1
        if window_left is None:
            window_left = window_dim - window_right - 1
        source, (in_spatial_dim,) = rf.pad(
            source,
            axes=[spatial_dim],
            padding=[(window_left, window_right)],
            value=pad_value,
        )
        # shape[0] == n_time + window - 1
    elif padding == "valid":
        in_spatial_dim = spatial_dim
        out_spatial_dim = spatial_dim - window_dim + 1
    else:
        raise ValueError(f"invalid padding {padding!r}")

    if stride > 1:
        start_times, out_spatial_dim = rf.range_over_dim_strided(out_spatial_dim, stride=stride)  # (n_out_time,)
        win_range = rf.range_over_dim(window_dim)  # (window,)
        indices = rf.combine_bc(start_times, "+", win_range)  # (n_out_time,window)
        final = rf.gather(source, indices=indices, axis=in_spatial_dim)  # (n_out_time,window,...)
        return final, out_spatial_dim

    tiled_dimshuffle = rf.expand_dim(source, dim=window_dim)  # (window,n_time+window-1,...)
    # We want to shift every dim*time block by one to the left.
    # To do this, we interpret that we have one more time frame (i.e. n_time+window).
    # We have to do some dimshuffling so that we get the right layout, then we can flatten,
    # add some padding, and then dimshuffle it back.
    # Then we can take out the first n_time frames.
    tiled_flat, flat_dim = rf.merge_dims(tiled_dimshuffle, dims=(window_dim, in_spatial_dim))
    rem = window_dim
    tiled_flat_pad_right, (flat_dim_ext,) = rf.pad(tiled_flat, axes=[flat_dim], padding=[(0, rem)], value=pad_value)
    # add time frame, (window,n_time+window,...)
    out_time_ext = out_spatial_dim + window_dim
    tiled_reshape_shift = rf.reshape(tiled_flat_pad_right, in_dims=[flat_dim_ext], out_dims=[window_dim, out_time_ext])
    final, _ = rf.slice(tiled_reshape_shift, axis=out_time_ext, size=out_spatial_dim)  # (window,n_time,...)
    if stride > 1:
        final, out_spatial_dim = rf.slice(final, axis=out_spatial_dim, step=stride)
    return final, out_spatial_dim


def concat(
    *sources: Tuple[Tensor, Dim],
    allow_broadcast: bool = False,
    out_dim: Optional[Dim] = None,
) -> Tuple[Tensor, Dim]:
    """
    Concatenates multiple sources in the specified dimension.

    Also see :func:`stack`.

    :param sources: list of (tensor, dim) pairs. dim is the axis to concatenate on.
    :param allow_broadcast: if True, the sources can have different dims, and the result will be broadcasted.
    :param out_dim: reuse existing dim for the resulting concatenated dim, if given
    :return: concatenated tensor, out_dim
    """
    assert sources
    if not allow_broadcast:
        dims = sources[0][0].dims_set - {sources[0][1]}
        for src, dim in sources:
            assert src.dims_set - {dim} == dims, f"concat {sources}, need allow_broadcast=True"
    if not out_dim:
        out_dim = sum(d for _, d in sources)
    # noinspection PyProtectedMember
    return sources[0][0]._raw_backend.concat(*sources, allow_broadcast=allow_broadcast, out_dim=out_dim), out_dim


def concat_features(*sources: Tensor, allow_broadcast=False) -> Tensor:
    """
    Concatenates multiple sources, using feature_dim of each source,
    so make sure that the feature_dim is correctly set.
    """
    src_pairs = []
    for src in sources:
        assert src.feature_dim is not None
        src_pairs.append((src, src.feature_dim))
    res, out_dim = concat(*src_pairs, allow_broadcast=allow_broadcast)
    res.feature_dim = out_dim
    return res


def pad(
    source: Tensor,
    *,
    axes: Sequence[Dim],
    padding: Sequence[Tuple[Union[Dim, int], Union[Dim, int]]],
    out_dims: Optional[Sequence[Dim]] = None,
    mode: str = "constant",
    value: Optional[Union[rf.RawTensorTypes, Tensor]] = None,
    handle_dynamic_dims: Optional[bool] = None,
) -> Tuple[Tensor, Sequence[Dim]]:
    """
    Pad values left/right in the specified axes.

    :param source:
    :param axes: which axes to add padding to
    :param padding: list of (left, right) padding for each axis
    :param out_dims: (optional) predefined out dims for each padded dim in axes. will automatically create if not given
    :param mode: 'constant', 'reflect', 'replicate' or 'circular'
    :param value: (optional) value to pad with in "constant" mode
    :param handle_dynamic_dims: True: when doing right padding on a dynamic dim, value will be added after the seq end,
        not at the end of the dimension. False: value will be added at the end of the dimension.
        By default, in behavior version >=21, this is True, in older versions, this is False.
    :return: padded tensor, out_dims. out dims are for each dim in axes
    """
    assert len(axes) == len(padding)
    if handle_dynamic_dims is None:
        handle_dynamic_dims = _pad_handle_dynamic_dims_default(axes, padding, mode=mode)
    if not out_dims:
        if handle_dynamic_dims:
            for left, right in padding:
                if isinstance(left, Dim):
                    assert not left.need_masking(), f"padding {padding} does not support dynamic left padding"
                if isinstance(right, Dim):
                    assert not right.need_masking(), f"padding {padding} does not support dynamic right padding"
                # Note that even dynamic middle dims is not exactly correct...
        out_dims = [left + middle + right for middle, (left, right) in zip(axes, padding)]
    # noinspection PyProtectedMember
    return (
        source._raw_backend.pad(
            source,
            axes=axes,
            padding=padding,
            out_dims=out_dims,
            handle_dynamic_dims=handle_dynamic_dims,
            mode=mode,
            value=value,
        ),
        out_dims,
    )


_pad_handle_dynamic_dims_shown_warning = False


def _pad_handle_dynamic_dims_default(
    pad_axes: Sequence[Dim], padding: Sequence[Tuple[Union[Dim, int], Union[Dim, int]]], *, mode: str
) -> bool:
    """
    :param pad_axes: list of axes to pad
    :param padding: list of (left, right) padding for each axis
    :param mode: 'constant', 'reflect', 'replicate' or 'circular'
    :return: True if dynamic dims should be handled as specified in the default behavior
    """
    from returnn.util.basic import BehaviorVersion

    if BehaviorVersion.get() >= 21:
        return True

    # Check whether not handling the dynamic dims is safe. Print a warning if not safe.
    global _pad_handle_dynamic_dims_shown_warning
    if not _pad_handle_dynamic_dims_shown_warning:
        for middle, (left, right) in zip(pad_axes, padding):
            middle: Dim
            if not middle.need_masking() and (isinstance(left, int) or not left.need_masking()):
                continue
            if mode != "circular" and isinstance(right, int) and right == 0:
                continue

            logging.getLogger("returnn.frontend").warning(
                f"rf.pad applied on dynamic dim {middle} but handle_dynamic_dims=False used by default"
                f" due to behavior version {BehaviorVersion.get()} < 21."
                " Set handle_dynamic_dims explicitly to avoid the warning,"
                " or switch to a new behavior version >= 21."
                " (This warning is only printed once.)"
            )
            _pad_handle_dynamic_dims_shown_warning = True
            break
    return False


def cum_concat_step(
    source: Tensor, *, prev_accum: Tensor, axis: Dim, out_spatial_dim: Optional[Dim] = None
) -> Tuple[Tensor, Dim]:
    """
    Concatenates all previous frames over a time-axis.
    See RETURNN :class:`CumConcatLayer` for details.

    :param source: same dims as prev_accum except for the accum axis
    :param prev_accum: previous accumulated tensor, shape {..., axis}
    :param axis: the axis to accumulate over
    :param out_spatial_dim: if given, the spatial dim of the output will be this dim. axis+1.
    :return: (accumulated, out_spatial_dim). accumulated shape {..., out_spatial_dim},
        same shape as prev_accum with axis replaced by out_spatial_dim.
    """
    if not out_spatial_dim:
        out_spatial_dim = axis + 1
    # noinspection PyProtectedMember
    return (
        source._raw_backend.cum_concat_step(source, prev_accum=prev_accum, axis=axis, out_spatial_dim=out_spatial_dim),
        out_spatial_dim,
    )


def stack(sources: Sequence[Tensor], *, out_dim: Optional[Dim] = None) -> Tuple[Tensor, Dim]:
    """
    Stack the sources in a new dimension.
    All sources must have the same shape.

    :param sources:
    :param out_dim: if given, use this as the new dim
    :return: stacked tensor, out_dim
    """
    if not sources:
        raise ValueError("no sources to stack")
    if not out_dim:
        out_dim = Dim(len(sources), name="stack")
    # noinspection PyProtectedMember
    return sources[0]._raw_backend.stack(sources, out_dim=out_dim), out_dim


def masked_select(
    tensor: Tensor, *, mask: Tensor, dims: Sequence[Dim], out_dim: Optional[Dim] = None
) -> Tuple[Tensor, Dim]:
    """
    This will pack the tensor based on the mask.
    In TF, this is ``boolean_mask``.
    The inverse of this is :func:`masked_scatter`.

    Related: :func:`pack_padded`, which uses :func:`sequence_mask` as the mask.

    :param tensor:
    :param mask:
    :param dims: The same or a subset of the dims of ``mask``.
        The order of the dims defines the packed format (in ``out_dim``).
        If only a single dim is given, only this dim is packed, while all other dims are kept as is.
    :param out_dim: packed ``dims`` into this dim. Automatically created if not given, and then returned.
    :return: Tensor where all ``dims`` are removed and replaced by a new dim (``out_dim``).
        The new dim (``out_dim``) is also returned.
        If mask==True for all elements, the returned tensor would be simply the flattened input tensor.
    """
    mask_dims_set = set(mask.dims)
    dims_set = set(dims)
    if not dims_set.issubset(mask_dims_set):
        raise ValueError(f"masked_select: dims {dims} not subset of mask dims {mask.dims}")
    if not dims_set:
        raise ValueError(f"masked_select: dims {dims} empty")
    if dims_set == mask_dims_set:
        # noinspection PyProtectedMember
        return tensor._raw_backend.masked_select(tensor, mask=mask, dims=dims, out_dim=out_dim)
    # Separate implementation for the case where we have a subset of the mask dims, specifically one single dim.
    # See https://github.com/rwth-i6/returnn/issues/1605 for discussion.
    mask = mask.copy_masked(mask_value=False, dims=dims)
    if len(dims) > 1:
        # Flatten it, in the specified order.
        tensor, in_dim = rf.merge_dims(tensor, dims=dims)
        mask, _ = rf.merge_dims(mask, dims=dims, out_dim=in_dim)
    else:
        (in_dim,) = dims
    in_dim: Dim
    idxs = rf.cumsum(rf.cast(mask, "int32"), spatial_dim=in_dim)  # [T,B] -> idx in T' + 1
    new_size = rf.gather(idxs, indices=in_dim.get_dim_value_tensor() - 1, axis=in_dim)  # [B]
    if out_dim is None:
        out_dim = Dim(new_size, name="masked_select")
    elif out_dim.dyn_size_ext is None:
        out_dim.dyn_size_ext = new_size
    elif out_dim.dyn_size_ext is not None and out_dim.dyn_size_ext.raw_tensor is None:
        out_dim.dyn_size_ext.raw_tensor = new_size.raw_tensor
    new_time = rf.reduce_max(new_size, axis=new_size.dims)  # T'
    idxs = rf.where(mask, idxs - 1, new_time)  # new_time is the padding idx
    ext_out_dim = out_dim + 1  # one more for the padded data
    ext_res = rf.scatter(tensor, indices=idxs, indices_dim=in_dim, out_dim=ext_out_dim)
    res, _ = rf.slice(ext_res, axis=ext_out_dim, size=out_dim)
    return res, out_dim


def masked_scatter(source: Tensor, *, mask: Tensor, dims: Sequence[Dim], in_dim: Dim) -> Tensor:
    """
    The inverse of :func:`masked_select`.

    :param source: [in_dim, F...]
    :param mask: [dims...] -> bool (e.g. [B,T])
    :param dims: the order of the dims defines the format. those dims should be exactly the dims of the mask.
    :param in_dim: the dim of the source which should be scattered into the mask.
    :return: [dims..., F...]
    """
    # noinspection PyProtectedMember
    return source._raw_backend.masked_scatter(source, mask=mask, dims=dims, in_dim=in_dim)


def sequence_mask(dims: Union[Dim, Sequence[Dim]], *, device: Optional[str] = None) -> Tensor:
    """
    :param dims:
    :param device:
    :return: mask based on the sequence lengths
    """
    if isinstance(dims, Dim):
        dims = [dims]
    assert len(dims) > 0
    dyn_dims = [d for d in dims if d.need_masking()]
    if not dyn_dims:
        return rf.constant(True, dims=())
    mask = True
    for dim in dyn_dims:
        mask = rf.opt_logical_and(mask, dim.get_mask(dim_order=dims, device=device))
    return mask


def pack_padded(
    source: Tensor, *, dims: Sequence[Dim], enforce_sorted: bool = False, out_dim: Optional[Dim] = None
) -> Tuple[Tensor, Dim]:
    """
    Like pack_padded_sequence. Usually the sequences are padded when they have different lengths.
    Packing means to only store the non-padded frames.
    This uses :func:`masked_select` internally based on the mask of non-masked frames.

    See :func:`pad_packed` for the inverse operation.

    :param source:
    :param dims: dims in source to pack. the order defines the format. first dim is major, etc.
        if there are no padded frames, e.g. dims=[B,T] would just result in the [B*T,...] reshaped tensor.
    :param enforce_sorted: seqs in the dims are reordered (stable sort) such that longest seqs come first.
    :param out_dim:
    :return: packed tensor, new packed dim
    """
    assert not enforce_sorted  # not implemented yet...
    mask = rf.sequence_mask(dims, device=source.device)
    # Note: We could already calculate out_dim here, as follows:
    #   out_dim = Dim(rf.num_elements_of_shape(dims), name="packed")
    # This could trigger a more efficient calculation path in masked_select,
    # where we can avoid a CUDA host-device synchronization, e.g. in the PyTorch backend.
    # However, in our benchmarks so far, it seems it's not helping so far,
    # so we don't do this, to also avoid the (minor) overhead of num_elements_of_shape here.
    # See: https://github.com/rwth-i6/returnn/pull/1593
    # This might change in the future when we have this:
    # https://github.com/pytorch/pytorch/issues/131256
    return rf.masked_select(source, mask=mask, dims=dims, out_dim=out_dim)


def pad_packed(source: Tensor, *, in_dim: Dim, dims: Sequence[Dim]) -> Tensor:
    """
    Inverse of :func:`pack_padded`, i.e. unpack the sequence, i.e. pad it back to the original length.
    """
    mask = rf.sequence_mask(dims, device=source.device)
    return rf.masked_scatter(source, mask=mask, in_dim=in_dim, dims=dims)


# noinspection PyUnusedLocal
def gather(
    source: Tensor,
    *,
    indices: Union[Tensor, int],
    axis: Optional[Dim] = None,
    clip_to_valid: bool = False,
) -> Tensor:
    """
    Gathers slices on a specified axis from the source using indices.
    If the source is of the shape ``[B,D,F1]``, and indices of shape ``[B,F2]``,
    this will yield output of the shape ``[B,F2,F1]`` where

    ``output[b,f2,f1] = source[b,indices[b,f2],f1]``

    (if ``D`` is the axis to gather from).
    In general, all shared axes of the input and the positions will be considered as batch-axes.

    The ``indices`` argument can also be an ``int``.
    In this case, this simply gives ``source[indices]`` on the specified ``axis``.

    :func:`scatter` is the inverse.

    :param source: [batch_dims..., axis, feature_dims...]
    :param indices: [batch_dims..., indices_dims...] indices used to select the slices of the source from.
        If another tensor, must be of type ``int32`` or ``int64``.
        Can also specify a constant ``int``.
        Batch dims are automatically determined as the common dims of source and indices.
    :param axis: The axis into which we gather the indices into.
        If not given, indices must be a tensor and the sparse_dim will be used.
    :param clip_to_valid: if True, the indices will be clipped to the valid range of the input
        Also taking seq lengths into account.
    :return: [batch_dims..., indices_dims..., feature_dims...] gathered values
    """
    if not axis:
        assert isinstance(indices, Tensor) and indices.sparse_dim
        axis = indices.sparse_dim
    # noinspection PyProtectedMember
    return source._raw_backend.gather(source, indices=indices, axis=axis, clip_to_valid=clip_to_valid)


def scatter(
    source: Tensor,
    *,
    indices: Tensor,
    indices_dim: Union[Dim, Sequence[Dim]],
    mode: str = "sum",
    fill_value: Optional[Union[int, float]] = None,
    out_dim: Optional[Union[Dim, Sequence[Dim]]] = None,
) -> Tensor:
    """
    Scatters into new zero-tensor.
    If entries in indices are duplicated, the corresponding values in source will be added together
    (scatter_add in PyTorch)
    with mode=="sum",
    or otherwise it will take the max/min.

    (TF segment_sum can be implemented via this.)

    :param source: [batch_dims..., indices_dim(s)..., feature_dims...]
    :param indices: [batch_dims..., indices_dim(s)...] -> out_dim
    :param indices_dim:
    :param mode: "sum" or "max" or "min". also see :func:`scatter_argmax`.
    :param fill_value:
    :param out_dim: The indices target dim.
        If not given, will be automatically determined as the sparse_dim from indices.
        If multiple out dims, use indices into the merged out dims,
        and then we use :func:`rf.split_dims` afterwards.
    :return: [batch_dims..., out_dim(s)..., feature_dims...]
    """
    if not out_dim:
        assert isinstance(indices, Tensor) and indices.sparse_dim
        out_dim = indices.sparse_dim
    if fill_value is None:
        if mode == "sum":
            fill_value = 0
        elif mode == "max":
            if "int" in source.dtype:
                import numpy

                fill_value = numpy.iinfo(source.raw_tensor.dtype).min
            else:
                fill_value = float("-inf")
        elif mode == "min":
            if "int" in source.dtype:
                import numpy

                fill_value = numpy.iinfo(source.raw_tensor.dtype).max
            else:
                fill_value = float("inf")
        else:
            raise ValueError(f"scatter: invalid mode {mode!r}")
    # noinspection PyProtectedMember
    return source._raw_backend.scatter(
        source, indices=indices, indices_dim=indices_dim, mode=mode, fill_value=fill_value, out_dim=out_dim
    )


def scatter_argmax(
    source: Tensor,
    *,
    indices: Tensor,
    indices_dim: Union[Dim, Sequence[Dim]],
    invalid_idx: int = -1,
    out_dim: Optional[Union[Dim, Sequence[Dim]]] = None,
) -> Tensor:
    """
    Get the index in src which has the max value for each index in index.

    This is like :func:`scatter` with ``mode="argmax"``.

    :param source: [batch_dims..., indices_dim(s)..., feature_dims...]
    :param indices: [batch_dims..., indices_dim(s)...] -> out_dim
    :param indices_dim:
    :param invalid_idx: in case some of the output entries are never set (via ``indices``),
        this will be used as the value.
    :param out_dim: The indices target dim.
    :return: [batch_dims..., out_dim(s)..., feature_dims...]
    """
    import numpy

    if not out_dim:
        assert isinstance(indices, Tensor) and indices.sparse_dim
        out_dim = indices.sparse_dim

    # For the shape comments, use [B,I,F] for shorter source, [B,O,F] for shorter output.
    # use scatter to get the max value for each index
    out_max = rf.scatter(source, indices=indices, indices_dim=indices_dim, mode="max", out_dim=out_dim)  # [B,O,F]
    src_max = rf.gather(out_max, indices=indices, axis=out_dim)  # [B,I,F] -> max value or invalid_value

    max_invalid_idx = numpy.iinfo(indices.dtype).max

    # then use gather to get the max value back to src.
    # then mask the src with the max value.
    src_max_mask = src_max == source
    src_max_mask = src_max_mask.copy_masked(False)
    src_indices = rf.where(
        src_max_mask, rf.range_over_dim(indices_dim, dtype=indices.dtype, device=source.device), max_invalid_idx
    )  # [B,I,F] -> I

    # now scatter the min of src_indices into tensor
    out = rf.scatter(
        src_indices, indices=indices, indices_dim=indices_dim, mode="min", fill_value=invalid_idx, out_dim=out_dim
    )  # [B,O,F] -> I or invalid_idx or max_invalid_idx

    if max_invalid_idx != invalid_idx:
        out = rf.where(out != max_invalid_idx, out, invalid_idx)  # [B,O,F] -> I or invalid_idx
    return out


# noinspection PyShadowingBuiltins
def slice(
    source: Tensor,
    *,
    axis: Dim,
    start: Optional[Union[int, Tensor]] = None,
    end: Optional[Union[int, Tensor]] = None,
    step: Optional[Union[int, Tensor]] = None,
    size: Optional[Union[int, Tensor, Dim]] = None,
    out_dim: Optional[Dim] = None,
) -> Tuple[Tensor, Dim]:
    """
    Slicing on the input, i.e. ``x[start:end:step]`` in some axis.

    If size is given, it takes out a slice-range like ``x[start:start + size]``.

    This function allows a non-scalar start points.

    :param source:
    :param axis:
    :param start:
    :param end:
    :param step: we only support a positive step here. If you want to flip the sequence,
        use :func:`reverse_sequence`, maybe with ``handle_dynamic_dims=False``.
    :param size:
    :param out_dim:
    :return: tensor, out_dim
    """
    if isinstance(step, int) and step <= 0:
        raise ValueError(f"rf.slice with non-positive step {step} not supported")
    if not out_dim:
        # Note: We cover not really all possible cases here but just those we really support.
        # (Actually, it might still be a superset of cases we really support,
        # but this might depend on the backend.)
        if size is not None:
            if isinstance(size, Dim):
                out_dim = size
            elif isinstance(size, (int, Tensor)):
                out_dim = Dim(size, name="slice")
            else:
                raise TypeError(f"invalid type {type(size)} for size {size}")
            assert step is None or (isinstance(step, int) and step == 1)

        else:  # size is None
            if start is None:
                start = 0
            if isinstance(start, int) and start >= 0:
                if end is None:
                    out_dim = axis.sub_left(start)
                elif isinstance(end, int):
                    if end < 0:
                        out_dim = axis.sub_left(start).sub_right(-end)
                    else:
                        out_dim = Dim(end - start, name="slice")
                elif isinstance(end, Tensor):
                    out_dim = Dim(end - start, name="slice")
                else:
                    raise TypeError(f"invalid type {type(end)} for end {end}")
            elif isinstance(start, int) and start < 0:
                if end is None:
                    out_dim = Dim(-start, name="slice")
                elif isinstance(end, int):
                    assert end < 0
                    out_dim = Dim(-start + end, name="slice")
                else:
                    raise TypeError(f"invalid type {type(end)} for end {end}")
            elif isinstance(start, Tensor):
                out_dim = Dim(
                    ((axis.get_size_tensor() if end is None else end) - start) % axis.get_size_tensor(), name="slice"
                )
            else:
                raise TypeError(f"invalid type {type(start)} for start {start}")

            if step is None or (isinstance(step, int) and step == 1):
                pass
            elif isinstance(step, int):
                out_dim = out_dim.ceildiv_right(step)
            elif isinstance(step, Tensor):
                step_dim = Dim(step, name="step")
                out_dim = out_dim.ceildiv_right(step_dim)
            else:
                raise TypeError(f"invalid type {type(step)} for step {step}")
    # noinspection PyProtectedMember
    return (
        source._raw_backend.slice(source, axis=axis, start=start, end=end, step=step, size=size, out_dim=out_dim),
        out_dim,
    )


def shift_right(source: Tensor, *, axis: Dim, pad_value: Union[rf.RawTensorTypes, Tensor], amount: int = 1) -> Tensor:
    """shift right by amount, pad left with left_pad"""
    padded, (padded_dim,) = rf.pad(source, axes=[axis], padding=[(amount, 0)], mode="constant", value=pad_value)
    padded_slice, _ = rf.slice(padded, axis=padded_dim, size=axis)
    return padded_slice


def shift_left(source: Tensor, *, axis: Dim, pad_value: Union[rf.RawTensorTypes, Tensor], amount: int = 1) -> Tensor:
    """shift left by amount, pad right with right_pad"""
    padded, (padded_dim,) = rf.pad(
        source, axes=[axis], padding=[(0, amount)], mode="constant", value=pad_value, handle_dynamic_dims=True
    )
    padded_slice, _ = rf.slice(padded, axis=padded_dim, start=amount, size=axis)
    return padded_slice


def reverse_sequence(tensor: Tensor, *, axis: Dim, handle_dynamic_dims: bool = True) -> Tensor:
    """
    Similar as tf.reverse_sequence, or Torch flip (but taking seq lengths into account).

    :param tensor:
    :param axis:
    :param handle_dynamic_dims: if True, the dynamic dims will be handled correctly,
        i.e. a padded seq "a b c <pad>" will be reversed to "c b a <pad>".
        If False, the dynamic dims will be handled as if they are static,
        i.e. a padded seq "a b c <pad>" will be reversed to "<pad> c b a".
    :return: reversed tensor, same dims
    """
    if not handle_dynamic_dims or not axis.need_masking():
        # noinspection PyProtectedMember
        return tensor._raw_backend.flip(tensor, axis=axis)
    indices = rf.combine_bc(axis.get_size_tensor(), "-", rf.range_over_dim(axis)) - 1
    return rf.gather(tensor, indices=indices, axis=axis, clip_to_valid=True)


def where(
    cond: Union[Tensor, rf.RawTensorTypes],
    true_: Union[Tensor, rf.RawTensorTypes],
    false_: Union[Tensor, rf.RawTensorTypes],
    *,
    allow_broadcast_all_sources: bool = False,
) -> Tensor:
    """
    Wraps tf.where, which is SwitchLayer in RETURNN.

    :return: true_ if cond else false_, elemwise.
    """
    if not isinstance(cond, Tensor):
        backend = _utils.get_backend_from_tensors(true_, false_)
        cond = rf.convert_to_tensor(cond, _backend=backend)
    # noinspection PyProtectedMember
    return cond._raw_backend.where(cond, true_, false_, allow_broadcast_all_sources=allow_broadcast_all_sources)


def search_sorted(
    sorted_seq: Tensor, values: Tensor, *, axis: Dim, side: str = "left", out_dtype: str = "int32"
) -> Tensor:
    """
    :param sorted_seq: [SharedDims...,axis], sequence of numbers, sorted low to high in the given axis.
    :param values: [SharedDims...,OtherDims...], sequence of numbers to search for in ``sorted_seq``.
    :param axis:
    :param side: "left" or "right"
    :param out_dtype:
    :return: [SharedDims...,OtherDims...] -> axis, indices in axis in ``sorted_seq`` such that
        sorted_seq[i-1] < value <= sorted_seq[i] if side=="left",
        sorted_seq[i-1] <= value < sorted_seq[i] if side=="right".
    """
    # noinspection PyProtectedMember
    return sorted_seq._raw_backend.search_sorted(sorted_seq, values, axis=axis, side=side, out_dtype=out_dtype)


def sparse_to_dense(
    labels: Union[Tensor, rf.RawTensorTypes],
    *,
    label_value: Union[Tensor, rf.RawTensorTypes],
    other_value: Union[Tensor, rf.RawTensorTypes],
    axis: Optional[Dim] = None,
) -> Tensor:
    """
    Converts a sparse tensor to a dense one.

    This is a more generic variant of "one_hot".

    Note that usually this is not needed as most other functions should handle sparse tensors just fine
    and much more efficiently than they would be with dense tensors.
    """
    labels = rf.convert_to_tensor(labels)
    if not axis:
        assert labels.sparse_dim, "sparse_to_dense: either provide `axis` or `labels` with sparse_dim"
        axis = labels.sparse_dim
    indices = rf.range_over_dim(axis)
    return where(rf.compare_bc(labels, "==", indices), label_value, other_value)


def one_hot(source: Tensor) -> Tensor:
    """
    one_hot. special case of :func:`sparse_to_dense`.

    Note that usually this is not needed as most other functions should handle sparse tensors just fine
    and much more efficiently than they would be with dense tensors.
    """
    return sparse_to_dense(source, label_value=1.0, other_value=0.0)



"""
Parameterizations using the parametrization API (:func:`register_parametrization`).

Also see:
https://github.com/rwth-i6/returnn/issues/1518
https://pytorch.org/tutorials/intermediate/parametrizations.html
"""

from __future__ import annotations
from returnn.tensor import Tensor
import returnn.frontend as rf


__all__ = ["weight_dropout", "WeightDropout", "weight_noise", "WeightNoise"]


def weight_dropout(module: rf.Module, param_name: str, *, drop_prob: float) -> rf.Module:
    """
    Apply weight dropout to a parameter of a module.

    This is only done in training.

    It uses :func:`gradient_checkpoint_scope` to avoid any memory overhead.

    In RETURNN TF-layers, this corresponds to the ``param_dropout`` option in a layer.
    Or in the RETURNN TF-layers :class:`RecLayer` with `ùnit="NativeLstm2"``,
    this was the ``rec_weight_dropout`` option.

    :param module:
    :param param_name: name of the parameter
    :param drop_prob: dropout probability
    :return: module
    """
    return rf.register_parametrization(module, param_name, WeightDropout(drop_prob))


class WeightDropout:
    """
    Use this for :func:`register_parametrization`, or via :func:`weight_dropout`.
    """

    def __init__(self, drop_prob: float):
        self.drop_prob = drop_prob

    def __call__(self, param: Tensor) -> Tensor:
        def _on_train() -> Tensor:
            with rf.gradient_checkpoint_scope():
                # on_forward=True because we already checked for train_flag
                return rf.dropout(param, drop_prob=self.drop_prob, on_forward=True)

        return rf.cond(rf.get_run_ctx().train_flag, _on_train, lambda: param)


def weight_noise(module: rf.Module, param_name: str, *, std: float) -> rf.Module:
    """
    Apply weight noise to a parameter of a module.
    This is also called variational noise.

    This is only done in training.

    It uses :func:`gradient_checkpoint_scope` to avoid any memory overhead.

    In RETURNN TF-layers, this corresponds to the ``param_variational_noise`` option in a layer.

    :param module:
    :param param_name: name of the parameter
    :param std: standard deviation of the noise
    :return: module
    """
    return rf.register_parametrization(module, param_name, WeightNoise(std))


class WeightNoise:
    """
    Use this for :func:`register_parametrization`, or via :func:`weight_noise`.
    """

    def __init__(self, std: float):
        self.std = std

    def __call__(self, param: Tensor) -> Tensor:
        def _on_train() -> Tensor:
            with rf.gradient_checkpoint_scope():
                noise = rf.random_normal(param.dims, dtype=param.dtype, stddev=self.std)
                return param + noise

        return rf.cond(rf.get_run_ctx().train_flag, _on_train, lambda: param)



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
import _setup_test_env  # noqa
import numpy as np
import returnn.frontend as rf
from returnn.tensor import Dim, TensorDict
from rf_utils import run_model


def test_piecewise_linear():
    batch_dim_ = Dim(13, name="batch")

    eps = 1e-5
    tests = np.array(
        [
            (0.0, 2.0),
            (1.0 - eps, 2.0),
            (1.0, 2.0),
            (1.0 + eps, 2.0),
            (2.0, 3.0),
            (3.0 - eps, 4.0),
            (3.0, 4.0),
            (3.0 + eps, 4.0),
            (4.0, 2.5),
            (5.0 - eps, 1.0),
            (5.0, 1.0),
            (5.0 + eps, 1.0),
            (6.0, 1.0),
        ],
        dtype="float32",
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.PiecewiseLinear, extern_data: TensorDict):
        values = rf.convert_to_tensor(tests[:, 0], dims=[batch_dim_])
        out = model(values)
        out.mark_as_default_output(shape=(batch_dim_,))

    out = run_model(
        TensorDict(),
        lambda *, epoch, step: rf.PiecewiseLinear({1: 2.0, 3: 4.0, 5: 1.0}),
        _forward_step,
        test_tensorflow=False,
    )
    out = out["output"]
    np.testing.assert_almost_equal(out.raw_tensor, tests[:, 1], decimal=4)



"""
Tensor binary operations mixin.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from .tensor import Tensor  # just for type hints; otherwise use _t.Tensor
    import returnn.frontend.types as _rf_types

from ._tensor_mixin_base import _TensorMixinBase


class _TensorOpOverloadsMixin(_TensorMixinBase):

    # Note that all those ops have native implementations as well,
    # so keep the logic in sync.

    # --- comparisons

    def __eq__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Union[Tensor, bool]:
        # When comparing to some other invalid type, return False, not a Tensor.
        # This is to allow easy equality checks with other random objects.
        # See for example here: https://github.com/rwth-i6/returnn/pull/1284
        if self.raw_tensor is None:
            # The other op overloads would actually raise some exception in this case.
            # However, here just return False.
            return False

        import returnn.frontend as rf

        valid_types = (rf.Tensor, self._raw_backend.RawTensorType) + tuple(rf.RawTensorTypes.__args__)
        if isinstance(other, valid_types):
            return _rf().compare(self, "==", other)
        return False

    def __ne__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().compare(self, "!=", other)

    def __lt__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().compare(self, "<", other)

    def __le__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().compare(self, "<=", other)

    def __gt__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().compare(self, ">", other)

    def __ge__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().compare(self, ">=", other)

    # --- math binary ops

    def __add__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(self, "+", other)

    def __radd__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(other, "+", self)

    def __sub__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(self, "-", other)

    def __rsub__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(other, "-", self)

    def __mul__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(self, "*", other)

    def __rmul__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(other, "*", self)

    def __truediv__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(self, "/", other)

    def __rtruediv__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(other, "/", self)

    def __floordiv__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(self, "//", other)

    def __rfloordiv__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(other, "//", self)

    def __mod__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(self, "%", other)

    def __rmod__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(other, "%", self)

    def __pow__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(self, "**", other)

    def __rpow__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(other, "**", self)

    def __and__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(self, "logical_and", other)

    def __rand__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(other, "logical_and", self)

    def __or__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(self, "logical_or", other)

    def __ror__(self: Tensor, other: Union[_rf_types.RawTensorTypes, Tensor]) -> Tensor:
        return _rf().combine(other, "logical_or", self)

    # --- math unary ops

    def __neg__(self: Tensor):  # -x
        return _rf().neg(self)

    def __invert__(self: Tensor):  # ~x, but for bool, treat is as logical_not, and otherwise not supported
        return _rf().logical_not(self)

    def __abs__(self: Tensor):
        return _rf().abs(self)

    def __ceil__(self: Tensor):
        return _rf().ceil(self)

    def __floor__(self: Tensor):
        return _rf().floor(self)


def _rf():
    import returnn.frontend as rf

    return rf



#!/usr/bin/env python3

"""
This is intended to test a compiled TF graph which was compiled via `compile_tf_graph.py`
with the option `--rec_step_by_step`.
This is just for demonstration, testing and debugging purpose. The search itself is performed in pure Python.
"""

# No RETURNN dependency needed for the basic search. Just TF itself.

import typing
import os
import json
import argparse
import tensorflow as tf
import numpy


class Hyp:
    """
    Represents a hypothesis in a given decoder step, including the label sequence so far.
    """

    def __init__(self, idx):
        """
        :param int idx: hyp idx (to identify it in a beam)
        """
        self.idx = idx
        self.source_idx = None  # type: typing.Optional[int]  # source hyp idx
        self.score = 0.0
        self.seq = []  # label seq

    def expand(self, idx, label, score):
        """
        :param int idx:
        :param int label:
        :param float score:
        :rtype: Hyp
        """
        new_hyp = Hyp(idx=idx)
        new_hyp.source_idx = self.idx
        new_hyp.seq = list(self.seq) + [label]
        new_hyp.score = score
        return new_hyp


def main():
    """
    Main entry.
    """
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--graph", help="compiled TF graph", required=True)
    arg_parser.add_argument("--chkpt", help="TF checkpoint (model params)", required=True)
    arg_parser.add_argument("--beam_size", type=int, default=12)
    arg_parser.add_argument("--rec_step_by_step_json", required=True)
    args = arg_parser.parse_args()

    # We operate only on a single seq (i.e. initially batch dim == 1, then batch dim == beam size).

    def make_initial_feed_dict():
        """
        :return: whatever placeholders we have for input features...
        :rtype: dict
        """
        # TODO...
        return {}

    # Load rec-step-by-step info. See compile_tf_graph.py for details.
    info = json.load(open(args.rec_step_by_step_json))
    assert isinstance(info, dict)

    # Load the graph
    if os.path.splitext(args.graph)[1] in [".meta", ".metatxt"]:  # meta graph
        saver = tf.compat.v1.train.import_meta_graph(args.graph)
    else:  # normal graph
        graph_def = tf.compat.v1.GraphDef()
        graph_def.ParseFromString(open(args.graph, "rb").read())
        tf.import_graph_def(graph_def)
        saver = tf.compat.v1.train.Saver()

    with tf.compat.v1.Session() as session:
        # Load the params.
        saver.restore(session, args.chkpt)

        # Calculate encoder states and initial decoder states.
        initial_feed_dict = make_initial_feed_dict()
        session.run(info["init_op"], feed_dict=initial_feed_dict)

        hyps = [Hyp(idx=0)]

        # Now loop over decoder steps.
        max_dec_len = 100  # TODO better default... depending on input len. or configurable...
        for i in range(max_dec_len):

            # Loop over all stochastic variables.
            for stochastic_var in info["stochastic_var_order"]:
                assert isinstance(stochastic_var, str)

                # Calculate the scores.
                session.run(info["stochastic_vars"][stochastic_var]["calc_scores_op"])
                # Get the scores (probabilities in +log space).
                scores = session.run(info["state_vars"]["stochastic_var_scores_%s" % stochastic_var])
                assert isinstance(scores, numpy.ndarray) and scores.ndim == 2 and scores.shape[0] == len(hyps)
                all_possibilities = [
                    (hyp.score + scores[i, j], j, hyp) for i, hyp in enumerate(hyps) for j in range(scores.shape[1])
                ]

                # TODO: length norm here?

                # Select new hypotheses.
                best_possibilities = sorted(all_possibilities)[
                    : args.beam_size
                ]  # type: typing.List[typing.Tuple[float,int,Hyp]]
                assert len(best_possibilities) == args.beam_size
                hyps = [
                    hyp.expand(idx=i, label=label, score=score)
                    for i, (score, label, hyp) in enumerate(best_possibilities)
                ]

                # Set the choices.
                session.run(
                    info["state_vars"]["stochastic_var_scores_%s" % stochastic_var] + "/Assign...?",  # TODO...
                    feed_dict={
                        info["state_vars"]["stochastic_var_scores_%s" % stochastic_var]
                        + "/Initial...?": [[hyp.seq[-1] for hyp in hyps]]  # TODO...
                    },
                )

                # Select source beams.
                session.run(
                    info["select_src_beams"]["op"],
                    feed_dict={info["select_src_beams"]["src_beams_placeholder"]: [[hyp.source_idx] for hyp in hyps]},
                )

            # Update state.
            session.run(info["next_step_op"])

            # TODO: stopping criterion?

    print("Best hypotheses:")
    for hyp in hyps:
        print("score %.2f: %r" % (hyp.score, hyp.seq))


if __name__ == "__main__":
    from returnn.util import better_exchook

    better_exchook.install()
    main()



from __future__ import annotations
from typing import List
import gc
import sys
import time
import numpy
import threading
import typing
from .basic import Dataset
from returnn.log import log
from returnn.util import NumbersDict


class CachedDataset(Dataset):
    """
    Base class for datasets with caching. This is only used for the :class:`HDFDataset`.
    Also see :class:`CachedDataset2`.
    """

    def __init__(self, cache_byte_size=0, **kwargs):
        """
        :param int cache_byte_size:
        """
        super(CachedDataset, self).__init__(**kwargs)
        self._cache_byte_size = cache_byte_size
        self.cache_byte_size_total_limit = cache_byte_size
        if cache_byte_size == -1:
            self.cache_byte_size_limit_at_start = 1024**4
        elif cache_byte_size == 0:
            self.cache_byte_size_limit_at_start = 0
        else:
            self.cache_byte_size_limit_at_start = max(cache_byte_size * 2 // 3, 1)
            self.cache_byte_size_total_limit = max(cache_byte_size - self.cache_byte_size_limit_at_start, 1)
        self.num_seqs_cached_at_start = 0
        self.cached_bytes_at_start = 0
        self.nbytes = 0
        self.start_cache_initialized = False
        self.definite_cache_leftover = 0
        self.cache_num_frames_free = 0
        self.preload_set = set([])
        self.preload_end = 0
        self.alloc_intervals = None  # type: typing.Optional[list]
        self._seq_start = []  # [numpy.array([0,0])]  # uses sorted seq idx, see set_batching()
        self._seq_index = []  # type: typing.List[int]  # Via init_seq_order(). seq_index idx -> hdf seq idx
        self._seq_index_inv = {}  # type: typing.Dict[int,int]  # Via init_seq_order(). hdf seq idx -> seq_index idx
        self._index_map = range(len(self._seq_index))  # sorted seq idx -> seq_index idx
        self._tag_idx = {}  # type: typing.Dict[str,int]  # map of tag -> real-seq-idx. call _update_tag_idx
        self.targets = {}
        self.target_keys = (
            []
        )  # the keys for which we provide data; we may have labels for additional keys in self.labels
        self.timestamps = None

    def initialize(self):
        """
        Initialization.
        """
        super(CachedDataset, self).initialize()

        if self.cache_byte_size_limit_at_start > 0:
            self.nbytes = numpy.array([], dtype=numpy.float32).itemsize * (self.num_inputs * self.window + 1 + 1)

            # Calculate cache sizes.
            temp_cache_size_bytes = max(0, self.cache_byte_size_total_limit)
            self.definite_cache_leftover = (
                temp_cache_size_bytes if self.num_seqs_cached_at_start == self.num_seqs else 0
            )
            self.cache_num_frames_free = temp_cache_size_bytes // self.nbytes

            print(
                "cached %i seqs" % self.num_seqs_cached_at_start,
                "%s GB" % (self.cached_bytes_at_start / float(1024 * 1024 * 1024)),
                ("(fully loaded, %s GB left over)" if self.definite_cache_leftover else "(%s GB free)")
                % max(temp_cache_size_bytes / float(1024 * 1024 * 1024), 0),
                file=log.v4,
            )

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        :type epoch: int|None
        :param list[str]|None seq_list: List of sequence tags, to set a predefined order.
        :param list[int]|None seq_order: List of corpus sequence indices, to set a predefined order.
        Initialize lists:
          self.seq_index  # sorted seq idx
        """
        super(CachedDataset, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        if seq_order is not None:
            seq_index = seq_order
        elif seq_list is not None:
            self._update_tag_idx()
            seq_index = [self._tag_idx[tag] for tag in seq_list]
        else:
            seq_index = self.get_seq_order_for_epoch(
                epoch, self._num_seqs, lambda s: self._get_seq_length_by_real_idx(s)[0]
            )

        old_index_map = self._index_map[:]
        self._index_map = range(len(seq_index))  # sorted seq idx -> seq_index idx

        if isinstance(seq_index, numpy.ndarray) or isinstance(self._seq_index, numpy.ndarray):
            seq_index_unchanged = numpy.array_equal(self._seq_index, seq_index)
        else:
            seq_index_unchanged = self._seq_index == seq_index

        if seq_index_unchanged and self.start_cache_initialized:
            return False

        if epoch is not None:
            # Give some hint to the user in case he is wondering why the cache is reloading.
            print("Reinitialize dataset seq order for epoch %i." % epoch, file=log.v4)

        if (
            self.cache_byte_size_limit_at_start == 0
            or self.num_seqs_cached_at_start != len(seq_index)
            or not self.start_cache_initialized
        ):
            self._seq_index = seq_index
            self._seq_index_inv = {}  # reset, create later if needed
            self._init_seq_starts()
            self._init_alloc_intervals()
            self._init_start_cache()
            self.start_cache_initialized = True
        else:
            if not self._seq_index_inv:
                self._seq_index_inv = dict(
                    zip(self._seq_index, range(len(self._seq_index)))
                )  # hdf seq idx -> seq_index idx
            self._index_map = [self._seq_index_inv[i] for i in seq_index]  # sorted seq idx -> seq_index idx
            if self._index_map == old_index_map:
                return False
        return True

    def supports_seq_order_sorting(self) -> bool:
        """supports sorting"""
        return True

    def supports_sharding(self) -> bool:
        """supports sharding"""
        return True

    def get_current_seq_order(self):
        assert self.cache_byte_size_limit_at_start == 0  # not implemented otherwise, we ignore _index_map
        return self._seq_index

    def _get_tag_by_real_idx(self, real_idx):
        raise NotImplementedError

    def _update_tag_idx(self):
        if self._tag_idx:
            return
        for i in range(self._num_seqs):
            self._tag_idx[self._get_tag_by_real_idx(i)] = i

    def batch_set_generator_cache_whole_epoch(self):
        return True

    def _init_alloc_intervals(self):
        if self.cache_byte_size_limit_at_start == 0:
            return
        if self.epoch is None:
            return
        assert self.num_seqs > 0
        assert self.num_inputs > 0
        assert self.window > 0
        self.preload_set = set([])
        self.alloc_intervals = [
            (0, 0, numpy.zeros([1] + self.get_data_shape("data"), dtype=self.get_data_dtype("data"))),
            (
                self.num_seqs,
                self.num_seqs,
                numpy.zeros([1] + self.get_data_shape("data"), dtype=self.get_data_dtype("data")),
            ),
        ]
        # self.alloc_intervals[i] is (idx start, idx end, data), where
        # idx start/end is the sorted seq idx start/end, end exclusive,
        # and data is a numpy.array.

    def _init_seq_starts(self):
        if self.cache_byte_size_limit_at_start == 0:
            return
        self._seq_start = [self._seq_start[0] * 0]  # idx like in seq_index, *not* real idx
        for i in range(self.num_seqs):
            ids = self._seq_index[i]
            self._seq_start.append(self._seq_start[-1] + self._get_seq_length_by_real_idx(ids))

    def _init_start_cache(self):
        if self.cache_byte_size_limit_at_start == 0:
            return
        if not self.alloc_intervals:
            return
        if not self.nbytes:
            return
        if not self.epoch:
            return

        num_cached = 0
        cached_bytes = 0
        for i in range(self.num_seqs):
            if i == num_cached:
                nbytes = self.get_seq_length_nd(i)[0] * self.nbytes
                if self.cache_byte_size_limit_at_start >= cached_bytes + nbytes:
                    num_cached = i + 1
                    cached_bytes += nbytes

        self.num_seqs_cached_at_start = num_cached
        self.cached_bytes_at_start = cached_bytes
        if num_cached > 0:
            self.preload_end = num_cached
            if sys.version_info >= (3, 0):
                threading.Thread(target=self._preload_seqs, args=(0, num_cached), daemon=True).start()
            else:
                threading.Thread(target=self._preload_seqs, args=(0, num_cached)).start()

    def load_seqs(self, start, end):
        """
        Load data sequences.
        As a side effect, will modify / fill-up:
          self.alloc_intervals
          self.targets
        This does some extra logic for the cache and calls self._load_seqs()
        for the real loading.

        :param int start: start sorted seq idx
        :param int end: end sorted seq idx
        """
        assert start >= 0
        assert start <= end

        if self.is_cached(start, end, blocking=False):
            return

        if self.cache_byte_size_limit_at_start > 0:  # If the cache is enabled.
            self._load_seqs_with_cache(start, end)
            return self.is_cached(start, end, blocking=True)

        super(CachedDataset, self).load_seqs(start, end)

    def _load_seqs(self, start, end):
        raise NotImplementedError

    def _load_seqs_with_cache(self, start, end, clear=True):
        if not clear:
            # only remove as many frames as required
            num_needed_cache_frames = self.get_seq_start(end)[0] - self.get_seq_start(start)[0]
            if self.cache_num_frames_free < num_needed_cache_frames:
                self.cache_num_frames_free += self.delete(num_needed_cache_frames - self.cache_num_frames_free)
                gc.collect()
            self.cache_num_frames_free -= num_needed_cache_frames
            threading.Thread(target=self._preload_seqs, args=(start, end)).start()
        else:
            # First, delete everything.
            self.cache_num_frames_free += self.delete(None)
            gc.collect()
            # Preload as much as we can so that we fill up the cache.
            while end < self.num_seqs:
                num_needed_cache_frames = self.get_seq_length_nd(end)[0]
                if self.cache_num_frames_free - num_needed_cache_frames < 0:
                    break
                self.cache_num_frames_free -= num_needed_cache_frames
                end += 1
            self.preload_end = end
            threading.Thread(target=self._preload_seqs, args=(start, end)).start()

    def _preload_seqs(self, start, end):
        print("Preloading cache from", start, "to", end, file=log.v4)
        super(CachedDataset, self).load_seqs(start, end)
        self.preload_end = self.num_seqs_cached_at_start

    def _shuffle_frames_in_seqs(self, start, end):
        """
        :type start: int
        :type end: int
        """
        assert start < end
        assert self.is_cached(start, end)
        alloc_idx = self.alloc_interval_index(start)
        alloc_start, alloc_end, alloc_data = self.alloc_intervals[alloc_idx]
        assert start >= alloc_start
        assert end <= alloc_end
        rnd = numpy.random.RandomState(start)  # Some deterministic way to shuffle!
        num_frames = self._seq_start[end][0] - self._seq_start[start][0]
        assert num_frames > 0
        perm = rnd.permutation(num_frames)
        alloc_offset = self._seq_start[start][0] - self._seq_start[alloc_start][0]
        assert alloc_offset + num_frames <= alloc_data.shape[0]
        # Permute alloc_data.
        data = alloc_data[alloc_offset : alloc_offset + num_frames]
        alloc_data[alloc_offset : alloc_offset + num_frames] = data[perm]
        # Permute targets.
        for k in self.targets:
            idx = self.target_keys.index(k) + 1
            targets = self.targets[k][self._seq_start[idx] : self._seq_start[start][idx] + num_frames]
            self.targets[k][
                self._seq_start[start][idx] : self._seq_start[start][idx]
                + self._seq_start[end][idx]
                - self._seq_start[start][idx]
            ] = targets[perm]

    def _set_alloc_intervals_data(self, idc, data):
        """
        :param int idc: index of sorted seq idx
        :param numpy.ndarray data: raw data
        """
        idi = self.alloc_interval_index(idc)
        assert idi >= 0
        o = self._seq_start[idc][0] - self._seq_start[self.alloc_intervals[idi][0]][0]
        l = data.shape[0]
        x = data
        if self.window > 1:
            x = self._sliding_window(x)
        self.alloc_intervals[idi][2][o : o + l] = x

    def alloc_interval_index(self, ids):
        """
        :param int ids: sorted seq idx
        :return index in self.alloc_intervals
        :rtype: int
        """
        s = 0
        e = len(self.alloc_intervals)
        # Binary search.
        while s < e:
            i = (s + e) // 2
            alloc_start, alloc_end, _ = self.alloc_intervals[i]
            if alloc_start <= ids < alloc_end:
                return i
            elif alloc_start <= ids and ids >= alloc_end:
                if s == i:
                    return -1
                s = i
            elif alloc_start > ids:
                if e == i:
                    return -1
                e = i
            else:
                assert False
        return -1

    def _insert_alloc_interval(self, pos, value, merge=False):
        """
        Insert np.zeros into self.alloc_intervals.
        :param int pos: idx in self.alloc_intervals
        :param (int,int) value: (start,end) like in load_seqs(), sorted seq idx
        :rtype: int
        """
        if value[0] == value[1]:
            return 0
        ci = self.alloc_intervals[pos][1]
        ni = self.alloc_intervals[pos + 1][0]
        xc = self.alloc_intervals[pos][2]
        xn = self.alloc_intervals[pos + 1][2]
        if value[0] == ci and value[1] == ni and merge:
            nj = self.alloc_intervals[pos][0]
            nk = self.alloc_intervals[pos + 1][1]
            del self.alloc_intervals[pos]
            del self.alloc_intervals[pos]
            self.alloc_intervals.insert(
                pos,
                (
                    nj,
                    nk,
                    numpy.concatenate(
                        [
                            xc,
                            numpy.zeros(
                                [self._seq_start[ni][0]] + self.get_data_shape("data"),
                                dtype=self.get_data_dtype("data"),
                            ),
                            xn,
                        ]
                    ),
                ),
            )
            return 0
        elif value[0] == ci and merge:
            nj = self.alloc_intervals[pos][0]
            del self.alloc_intervals[pos]
            self.alloc_intervals.insert(
                pos,
                (
                    nj,
                    value[1],
                    numpy.concatenate(
                        [
                            xc,
                            numpy.zeros(
                                [self._seq_start[value[1]][0] - self._seq_start[ci][0]] + self.get_data_shape("data"),
                                dtype=self.get_data_dtype("data"),
                            ),
                        ]
                    ),
                ),
            )
            return 0
        elif value[1] == ni and merge:
            nk = self.alloc_intervals[pos + 1][1]
            del self.alloc_intervals[pos + 1]
            self.alloc_intervals.insert(
                pos + 1,
                (
                    value[0],
                    nk,
                    numpy.concatenate(
                        [
                            numpy.zeros(
                                [self._seq_start[ni][0] - self._seq_start[value[0]][0]] + self.get_data_shape("data"),
                                dtype=self.get_data_dtype("data"),
                            ),
                            xc,
                        ]
                    ),
                ),
            )
            return 0
        else:
            self.alloc_intervals.insert(
                pos + 1,
                value
                + (
                    numpy.zeros(
                        [self._seq_start[value[1]][0] - self._seq_start[value[0]][0]] + self.get_data_shape("data"),
                        dtype=self.get_data_dtype("data"),
                    ),
                ),
            )
            return 1

    def _remove_alloc_interval(self, pos, value):
        """
        Remove data from self.alloc_intervals.
        :param int pos: idx in self.alloc_intervals
        :param (int,int) value: (start,end) like in load_seqs(), sorted seq idx
        :rtype: int
        """
        ci, ni, xi = self.alloc_intervals[pos]
        if value[0] == ci and value[1] == ni:
            del self.alloc_intervals[pos]
            return -1
        elif value[0] == ci:
            self.alloc_intervals.insert(
                pos, (value[1], ni, xi[self._seq_start[value[1]][0] - self._seq_start[ci][0] :])
            )
            del self.alloc_intervals[pos + 1]
            return 0
        elif value[1] == ni:
            self.alloc_intervals.insert(
                pos, (ci, value[0], xi[: self._seq_start[value[0]][0] - self._seq_start[ci][0]])
            )
            del self.alloc_intervals[pos + 1]
            return 0
        else:
            self.alloc_intervals.insert(
                pos, (value[1], ni, xi[self._seq_start[value[1]][0] - self._seq_start[ci][0] :])
            )
            self.alloc_intervals.insert(
                pos, (ci, value[0], xi[: self._seq_start[value[0]][0] - self._seq_start[ci][0]])
            )
            del self.alloc_intervals[pos + 2]
            return 1

    def _modify_alloc_intervals(self, start, end, invert):
        """
        Inserts/removes sorted seq idx range (start,end).
        :param int start: like in load_seqs(), sorted seq idx
        :param int end: like in load_seqs(), sorted seq idx
        :param bool invert: True->insert, False->remove
        :rtype: list[int]
        :return selection list, modified sorted seq idx in self.alloc_intervals
        """
        if end is None:
            end = start + 1
        if start == end:
            return
        assert start < end
        i = 0
        selection = []  # type: typing.List[int]
        modify = self._insert_alloc_interval if invert else self._remove_alloc_interval
        while i < len(self.alloc_intervals) - invert:
            ni = self.alloc_intervals[i + invert][1 - invert]  # insert mode: start idx of next alloc
            ci = self.alloc_intervals[i][invert]  # insert mode: end idx of cur alloc
            assert ci <= ni
            flag = ((ci <= start < ni), (ci < end <= ni), (ci < start and ni <= start) or (ci >= end and ni > end))
            if not flag[0] and not flag[1]:
                if not flag[2]:
                    selection.extend(range(ci, ni))
                    i += modify(i, (ci, ni))
            elif flag[1]:
                v = (start if flag[0] else ci, end)
                selection.extend(range(v[0], v[1]))
                i += modify(i, v)
                break
            elif flag[0]:
                selection.extend(range(start, ni))
                i += modify(i, (start, ni))
            i += 1
        if self.alloc_intervals[0][0] != 0:
            self.alloc_intervals.insert(
                0, (0, 0, numpy.zeros([1] + self.get_data_shape("data"), dtype=self.get_data_dtype("data")))
            )
        if self.alloc_intervals[-1][1] != self.num_seqs:
            self.alloc_intervals.append(
                (
                    self.num_seqs,
                    self.num_seqs,
                    numpy.zeros([1] + self.get_data_shape("data"), dtype=self.get_data_dtype("data")),
                )
            )
        return selection

    def insert_alloc_interval(self, start, end=None):
        return self._modify_alloc_intervals(start, end, True)

    def remove_alloc_interval(self, start, end=None):
        return self._modify_alloc_intervals(start, end, False)

    def delete(self, nframes):
        """
        :param int|None nframes: how much frames to delete max.
          Note that this limit is not strict. We can end up
          deleting more than nframes.
        :return: number of frames deleted
        :rtype: int
        """
        if nframes is not None:
            if nframes == 0:
                return 0
            assert nframes > 0
        deleted = 0
        i = 0
        while (not nframes or deleted < nframes) and i < len(self.alloc_intervals):
            ai = self.alloc_intervals[i]
            if ai[1] > self.num_seqs_cached_at_start and ai[0] < ai[1]:
                removed = self.remove_alloc_interval(max(ai[0], self.num_seqs_cached_at_start), ai[1])
                self.preload_set -= set(removed)
                deleted += sum([self._get_seq_length_by_real_idx(self._seq_index[i])[0] for i in removed])
            else:
                i += 1
        return deleted

    @property
    def num_seqs(self):
        return len(self._index_map)

    def is_cached(self, start, end, blocking=False):
        """
        :param int start: like in load_seqs(), sorted seq idx
        :param int end: like in load_seqs(), sorted seq idx
        :rtype: bool
        :returns whether we have the full range (start,end) of sorted seq idx
          cached in self.alloc_intervals (end is exclusive).
        """
        if self.cache_byte_size_total_limit == 0:  # disabled cache
            return False
        if start == end:
            return True  # Empty.
        assert start < end
        if blocking and end <= self.preload_end:
            while not set(range(start, end)) <= self.preload_set:
                time.sleep(0.2)
            return True
        return set(range(start, end)) <= self.preload_set

    def _get_seq_length_by_real_idx(self, real_seq_idx):
        """
        :param int real_seq_idx:
        :returns length of the sequence with index 'real_seq_idx'
        :rtype: numpy.ndarray
        """
        raise NotImplementedError

    def get_seq_length_nd(self, sorted_seq_idx):
        """
        :type sorted_seq_idx: int
        :rtype: numpy.ndarray
        """
        real_seq_idx = self._seq_index[self._index_map[sorted_seq_idx]]
        return self._get_seq_length_by_real_idx(real_seq_idx)

    def get_seq_length(self, seq_idx):
        """
        :rtype: NumbersDict
        """
        lengths = self.get_seq_length_nd(seq_idx)
        d = {}
        first_target_idx = 0
        # We allow using only targets. In this case self.num_inputs == 0 and the "data" key is not used.
        if self.num_inputs > 0:
            d["data"] = lengths[0]
            first_target_idx = 1
        for k, l in zip(self.target_keys, lengths[first_target_idx:]):
            d[k] = l
        return NumbersDict(d)

    def get_seq_start(self, sorted_seq_idx):
        """
        :type sorted_seq_idx: int
        :rtype: (int,int)
        """
        return self._seq_start[sorted_seq_idx]

    def get_times(self, sorted_seq_idx):
        seq_start = self.get_seq_start(sorted_seq_idx)[0]
        seq_len = self.get_seq_length_nd(sorted_seq_idx)[0]
        return self.timestamps[seq_start : seq_start + seq_len]

    def get_input_data(self, sorted_seq_idx):
        seq_idx = self._index_map[sorted_seq_idx]
        idi = self.alloc_interval_index(seq_idx)
        assert idi >= 0, "failed to get data for seq %i" % sorted_seq_idx
        alloc_start_seq, alloc_end_seq, alloc_data = self.alloc_intervals[idi]
        o = self.get_seq_start(seq_idx)[0] - self.get_seq_start(alloc_start_seq)[0]
        assert o >= 0
        l = self.get_seq_length_nd(sorted_seq_idx)[0]
        assert alloc_data.shape[0] >= o + l
        return alloc_data[o : o + l]

    def get_data_dim(self, key):
        if key == "data" and self.num_inputs > 0:  # if num_inputs == 0, we allow "data" as a target key
            return self.num_inputs * self.window
        return self.num_outputs[key][0]

    def get_targets(self, target, sorted_seq_idx):
        seq_idx = self._index_map[sorted_seq_idx]
        idx = self.target_keys.index(target) + 1
        seq_start = self.get_seq_start(seq_idx)[idx]
        seq_len = self.get_seq_length_nd(sorted_seq_idx)[idx]
        return self.targets[target][seq_start : seq_start + seq_len]

    def get_data_keys(self) -> List[str]:
        """data keys"""
        res = []
        if self.num_inputs > 0:
            res.append("data")
        res.extend(self.get_target_list())
        return res

    def get_target_list(self):
        return list(self.targets.keys())

    def get_tag(self, sorted_seq_idx):
        raise NotImplementedError

    def have_corpus_seq_idx(self):
        return True

    def get_corpus_seq_idx(self, seq_idx):
        """
        :param int seq_idx: sorted sequence index from the current epoch, depending on seq_ordering
        :return: the sequence index as-is in the original corpus. only defined if self.have_corpus_seq_idx()
        :rtype: int
        """
        return self._seq_index[self._index_map[seq_idx]]



"""
Utilities for PyTorch tests
"""

from __future__ import annotations
from typing import Optional, Any, Sequence, Tuple, List, Dict
import torch


def report_profile(
    prof: torch.profiler.profiler,
    check_events: Optional[Sequence[Tuple[str, Dict[str, Any]]]] = None,
    *,
    allow_remaining_allocs: bool = False,
    _size_threshold: int = 100,
) -> None:
    """
    Report profile.

    Example usage::

        from torch.profiler import profile, ProfilerActivity

        with profile(
            activities=[ProfilerActivity.CPU], profile_memory=True, with_stack=True, record_shapes=True
        ) as prof:
            # code to profile
            ...

        report_profile(prof)

    :param prof: via torch.profiler.profile.
    :param check_events: if given, will check that the report matches the given events.
        Each entry is a tuple (event_name, event_opts).
        The event_name can be "alloc", "dealloc", "torchop", "pycall".
        The event_opts is a dict with the expected values.
        You can use "*" as a wildcard in the event_opts.
    :param allow_remaining_allocs: if True, will not raise an error if there are remaining allocations.
    :param _size_threshold: internal
    """
    # Note: I tried prof.events(), prof.profiler.kineto_results.events(), prof._memory_profile().timeline,
    # but they all are not really giving me the information I want.
    # Either the Python stack is missing, or the memory information is incomplete,
    # or the Python/TorchOp events are missing.
    # The only complete information source seems to be prof.profiler.kineto_results.experimental_event_tree().

    import fnmatch

    # noinspection PyProtectedMember
    from torch.profiler._utils import traverse_dfs
    from torch._C._profiler import _EventType  # noqa

    _printed_events = set()
    _events_stack_depth = {}
    _allocs = {}  # id -> dict with "size", "name"
    check_events = list(check_events) if check_events is not None else None

    def _ev_visit(ev):
        # ev: torch._C._profiler._ProfilerEvent
        if ev.typed[0] == _EventType.Allocation:
            ex = ev.typed[1]  # torch._C._profiler._ExtraFields_Allocation
            # ex.id/ex.allocation_id/ex.ptr redundant?
            if ex.allocation_id in _allocs:
                ev_name = "dealloc"  # deallocation
                assert _allocs[ex.allocation_id]["size"] == -ex.alloc_size
                name = _allocs[ex.allocation_id]["name"]
                del _allocs[ex.allocation_id]
            else:
                ev_name = "alloc"
                assert ex.alloc_size > 0
                assert ev.parent
                name = _ev_ctx(ev.parent)
                _allocs[ex.allocation_id] = {"size": ex.alloc_size, "name": name}
            opts = {"id": ex.allocation_id, "name": name, "size": ex.alloc_size, "total_alloc": ex.total_allocated}
        elif ev.typed[0] == _EventType.TorchOp:
            ev_name = "torchop"
            ex = ev.typed[1]  # torch._C._profiler._ExtraFields_TorchOp
            opts = {"name": ex.name, "inputs": [_repr_tensor_metadata(i) for i in ex.inputs]}
        elif ev.typed[0] == _EventType.PyCall:
            ev_name = "pycall"
            ex = ev.typed[1]  # torch._C._profiler._ExtraFields_PyCall
            ex0 = ex.caller  # torch._C._profiler._PyFrameState
            ex1 = ex.callsite  # torch._C._profiler._PyFrameState
            if _pycall_filter_fn(ex0.file_name) or _pycall_filter_fn(ex1.file_name):
                opts = {
                    "caller_loc": f"{ex0.file_name}:{ex0.line_number}",
                    "caller_name": ex0.function_name,
                    "callsite_name": ex1.function_name,
                }
            else:
                return
        elif ev.typed[0] == _EventType.PyCCall:
            ev_name = "pyccall"
            ex = ev.typed[1]  # torch._C._profiler._ExtraFields_PyCCall
            ex0 = ex.caller  # torch._C._profiler._PyFrameState
            if _pycall_filter_fn(ex0.file_name):
                opts = {
                    "caller_loc": f"{ex0.file_name}:{ex0.line_number}",
                    "caller_name": ex0.function_name,
                }
            else:
                return
        else:
            return

        depth = 0
        if ev.parent:
            ev__ = ev.parent
            while ev__:
                if ev__ in _printed_events:
                    depth = _events_stack_depth[ev__] + 1
                    break
                if ev__ in _events_stack_depth:
                    depth = _events_stack_depth[ev__]
                    break
                ev__ = ev__.parent
            ev__ = ev
            while ev__:
                if ev__ in _events_stack_depth:
                    break
                _events_stack_depth[ev__] = depth
                ev__ = ev__.parent
        else:
            _events_stack_depth[ev] = 0
        _printed_events.add(ev)
        prefix = "  " * depth

        if check_events is None:
            print(f"{prefix}{ev_name} {opts}")
            return

        next_check = check_events[0] if check_events else None
        if next_check:
            next_check_name, next_check_opts = next_check
            if ev_name == next_check_name:
                for k, v in next_check_opts.items():
                    if isinstance(v, str) and "*" in v:
                        if not fnmatch.fnmatch(opts[k], v):
                            mismatch = f"Pattern mismatch: {opts[k]} vs {v}"
                            break
                    elif k == "total_alloc":
                        if abs(opts[k] - v) >= _size_threshold:
                            mismatch = f"Size mismatch: {opts[k]} vs {v}"
                            break
                    elif opts[k] != v:
                        mismatch = f"Value mismatch: {opts[k]} vs {v}"
                        break
                else:
                    mismatch = None
            else:
                mismatch = f"Different event: {ev_name} vs {next_check_name}"
        else:
            mismatch = "No check event"

        if ev_name in {"alloc", "dealloc"} and abs(opts["size"]) >= _size_threshold:
            assert not mismatch, f"Event not matched: {ev_name} {opts} to {next_check}: {mismatch}"

        if not mismatch:
            print(f"{prefix}{ev_name} {opts} ✓")
            check_events.pop(0)
        else:
            print(f"{prefix}({ev_name} {opts})")

    for ev_ in sorted(
        traverse_dfs(prof.profiler.kineto_results.experimental_event_tree()), key=lambda ev: ev.start_time_ns
    ):
        # ev: torch._C._profiler._ProfilerEvent
        _ev_visit(ev_)

    if allow_remaining_allocs:
        for alloc_id, alloc in _allocs.items():
            print(f"Remaining alloc: {alloc_id} {alloc}")
    else:
        assert not _allocs, f"Remaining allocs: {_allocs}"
    assert not check_events, f"Remaining check events: {check_events}"


def get_remaining_allocs_from_profile(prof: torch.profiler.profiler) -> Dict[int, Dict[str, Any]]:
    """
    Get remaining allocs from profile.

    :param prof: via torch.profiler.profile.
    :return: allocs dict: id -> dict with "size", "name"
    """
    # noinspection PyProtectedMember
    from torch.profiler._utils import traverse_dfs
    from torch._C._profiler import _EventType  # noqa

    _allocs = {}  # id -> dict with "size", "name"

    for ev in sorted(
        traverse_dfs(prof.profiler.kineto_results.experimental_event_tree()), key=lambda ev: ev.start_time_ns
    ):
        # ev: torch._C._profiler._ProfilerEvent
        if ev.typed[0] == _EventType.Allocation:
            ex = ev.typed[1]  # torch._C._profiler._ExtraFields_Allocation
            # ex.id/ex.allocation_id/ex.ptr redundant?
            if ex.allocation_id in _allocs:
                # expect deallocation
                assert _allocs[ex.allocation_id]["size"] == -ex.alloc_size
                del _allocs[ex.allocation_id]
            else:
                # allocation
                assert ex.alloc_size > 0
                assert ev.parent
                name = _ev_ctx(ev.parent)
                _allocs[ex.allocation_id] = {"size": ex.alloc_size, "name": name}

    return _allocs


def get_allocs_from_profile(prof: torch.profiler.profiler) -> List[Dict[str, Any]]:
    """
    Get allocs from profile.

    :param prof: via torch.profiler.profile.
    :return: allocs dict: id -> dict with "size", "name"
    """
    # noinspection PyProtectedMember
    from torch.profiler._utils import traverse_dfs
    from torch._C._profiler import _EventType  # noqa

    _allocs = []  # dict with "id", "size", "name"

    for ev in sorted(
        traverse_dfs(prof.profiler.kineto_results.experimental_event_tree()), key=lambda ev: ev.start_time_ns
    ):
        # ev: torch._C._profiler._ProfilerEvent
        if ev.typed[0] == _EventType.Allocation:
            ex = ev.typed[1]  # torch._C._profiler._ExtraFields_Allocation
            if ex.alloc_size > 0:
                assert ev.parent
                name = _ev_ctx(ev.parent)
                _allocs.append({"id": ex.allocation_id, "size": ex.alloc_size, "name": name})

    return _allocs


def get_peak_alloc_from_profile(prof: torch.profiler.profiler) -> int:
    """
    Get remaining allocs from profile.

    :param prof: via torch.profiler.profile.
    :return: peak alloc size
    """
    # noinspection PyProtectedMember
    from torch.profiler._utils import traverse_dfs
    from torch._C._profiler import _EventType  # noqa

    _allocs = {}  # id -> dict with "size", "name"
    peak_alloc = 0

    for ev in sorted(
        traverse_dfs(prof.profiler.kineto_results.experimental_event_tree()), key=lambda ev: ev.start_time_ns
    ):
        # ev: torch._C._profiler._ProfilerEvent
        # ev: torch._C._profiler._ProfilerEvent
        if ev.typed[0] == _EventType.Allocation:
            ex = ev.typed[1]  # torch._C._profiler._ExtraFields_Allocation
            # ex.id/ex.allocation_id/ex.ptr redundant?
            if ex.allocation_id in _allocs:
                # expect deallocation
                assert _allocs[ex.allocation_id]["size"] == -ex.alloc_size
                del _allocs[ex.allocation_id]
            else:
                # allocation
                assert ex.alloc_size > 0
                assert ev.parent
                name = _ev_ctx(ev.parent)
                _allocs[ex.allocation_id] = {"size": ex.alloc_size, "name": name}

                cur_total_alloc = sum(alloc["size"] for alloc in _allocs.values())
                if cur_total_alloc > peak_alloc:
                    peak_alloc = cur_total_alloc

    return peak_alloc


def _pycall_filter_fn(filename: str) -> bool:
    assert not filename.startswith("/")  # currently the case...
    if filename.startswith("test_"):
        assert "/" not in filename  # currently the case...
        return True
    if filename.startswith("returnn/"):
        return True
    return False


def _ev_ctx(ev) -> str:
    from torch._C._profiler import _EventType  # noqa

    stack = [None]
    parent = ev
    while parent and parent.typed[0] == _EventType.TorchOp:  # go to top torch op
        stack[-1] = parent.typed[1].name
        parent = parent.parent
    if not stack[-1] and parent.typed[0] == _EventType.PyCCall:
        stack[-1] = parent.typed[1].caller.function_name
        parent = parent.parent
    if not stack[-1]:
        stack.pop(-1)
    while parent:
        if parent.typed[0] == _EventType.PyCall:
            ex0 = parent.typed[1].caller  # torch._C._profiler._PyFrameState
            ex1 = parent.typed[1].callsite  # torch._C._profiler._PyFrameState
            if (
                _pycall_filter_fn(ex1.file_name)
                or (_pycall_filter_fn(ex0.file_name) and ex1.function_name == "backward")
            ) and ex1.function_name not in {"__torch_dispatch__"}:
                stack.append(ex1.function_name)
        parent = parent.parent
    stack.reverse()
    return "/".join(stack) or "unknown"


def _repr_tensor_metadata(x) -> Any:
    """
    :param x: torch._C._profiler._TensorMetadata or int
    :return: obj which has a nice __repr__/__str__
    """
    if isinstance(x, (list, tuple)):
        return type(x)(_repr_tensor_metadata(i) for i in x)
    if not type(x).__module__.startswith("torch"):
        return x
    assert x is not None
    return _ReprObj(f"<_T id={x.id} shape={x.sizes} dtype={x.dtype}>")


class _ReprObj:
    def __init__(self, obj: str):
        self.obj = obj

    def __repr__(self):
        return self.obj



"""
Vocabulary related classes for targets such as BPE, SentencePieces etc...
"""

from __future__ import annotations

__all__ = [
    "Vocabulary",
    "BytePairEncoding",
    "SamplingBytePairEncoding",
    "SentencePieces",
    "CharacterTargets",
    "Utf8ByteTargets",
]

from typing import Optional, Union, Type, List
import typing
import sys
import numpy

from returnn.log import log
from returnn.util.basic import NotSpecified


class Vocabulary:
    """
    Represents a vocabulary (set of words, and their ids).
    Used by :class:`BytePairEncoding`.
    """

    _cache = {}  # filename -> vocab dict, labels dict (see _parse_vocab)

    @classmethod
    def create_vocab(cls, **opts):
        """
        :param opts: kwargs for class
        :rtype: Vocabulary|BytePairEncoding|CharacterTargets
        """
        opts = opts.copy()
        clz = cls
        if "class" in opts:
            class_name = opts.pop("class")
            clz = globals()[class_name]
            assert issubclass(clz, Vocabulary), "class %r %r is not a subclass of %r" % (class_name, clz, cls)
        elif "bpe_file" in opts:
            clz = BytePairEncoding
        return clz(**opts)

    def __init__(
        self,
        vocab_file,
        seq_postfix=None,
        unknown_label="UNK",
        bos_label=None,
        eos_label=None,
        pad_label=None,
        control_symbols=None,
        user_defined_symbols=None,
        num_labels=None,
        labels=None,
    ):
        """
        :param str|None vocab_file:
        :param str|int|None unknown_label: e.g. "UNK" or "<unk>"
        :param str|int|None bos_label: e.g. "<s>"
        :param str|int|None eos_label: e.g. "</s>"
        :param str|int|None pad_label: e.g. "<pad>"
        :param dict[str,str|int]|None control_symbols:
          https://github.com/google/sentencepiece/blob/master/doc/special_symbols.md
        :param dict[str,str|int]|None user_defined_symbols:
          https://github.com/google/sentencepiece/blob/master/doc/special_symbols.md
        :param int num_labels: just for verification
        :param list[int]|None seq_postfix: labels will be added to the seq in self.get_seq
        :param list[str]|(()->list[str])|None labels:
        """
        self.vocab_file = vocab_file
        self.unknown_label = unknown_label
        self.num_labels = None  # type: typing.Optional[int]  # will be set by _parse_vocab
        self._vocab = None  # type: typing.Optional[typing.Dict[str,int]]  # label->idx
        if labels is not None and callable(labels):
            labels = labels()
        if labels is not None:
            assert isinstance(labels, (list, tuple))
        self._labels = labels

        self._parse_vocab()
        if num_labels is not None:
            assert self.num_labels == num_labels
        self.unknown_label_id = self.to_id(self.unknown_label, allow_none=True)
        if self.unknown_label_id is not None:
            self.unknown_label = self.id_to_label(self.unknown_label_id)
        self.bos_label_id = self.to_id(bos_label, allow_none=True)
        self.eos_label_id = self.to_id(eos_label, allow_none=True)
        self.pad_label_id = self.to_id(pad_label, allow_none=True)
        self.control_symbol_ids = {name: self.to_id(label) for name, label in (control_symbols or {}).items()}
        self.user_defined_symbol_ids = {name: self.to_id(label) for name, label in (user_defined_symbols or {}).items()}
        self.seq_postfix = seq_postfix or []

    def __repr__(self):
        parts = [repr(self.vocab_file), "num_labels=%s" % self.num_labels]
        if self.unknown_label_id is not None:
            parts.append("unknown_label=%r" % self.unknown_label)
        if self.bos_label_id is not None:
            parts.append("bos_label=%r" % self.id_to_label(self.bos_label_id))
        if self.eos_label_id is not None:
            parts.append("eos_label=%r" % self.id_to_label(self.eos_label_id))
        if self.pad_label_id is not None:
            parts.append("pad_label=%r" % self.id_to_label(self.pad_label_id))
        return "%s(%s)" % (self.__class__.__name__, ", ".join(parts))

    def set_random_seed(self, seed: int):
        """
        This can be called for a new epoch or so.
        Usually it has no effect, as there is no randomness.
        However, some vocab class could introduce some sampling process.

        :param seed:
        """
        pass  # usually there is no randomness, so ignore

    def _parse_vocab(self):
        """
        Sets self.vocab, self.labels, self.num_labels.
        """
        filename = self.vocab_file
        import pickle

        if self._labels is not None:
            self._vocab = {label: i for i, label in enumerate(self._labels)}
            self.num_labels = len(self._labels)
        elif filename in self._cache:
            self._vocab, self._labels = self._cache[filename]
            self.num_labels = len(self._labels)
        else:
            if filename[-4:] == ".pkl":
                d = pickle.load(open(filename, "rb"))
            else:
                file_content = open(filename, "r").read()
                if file_content.startswith("{"):
                    d = eval(file_content)
                else:
                    # Do line-based parsing.
                    lines = file_content.splitlines()
                    d = {line: i for (i, line) in enumerate(lines)}
            assert isinstance(d, dict), f"{self}: expected dict, got {type(d).__name__} in {filename}"
            labels = {idx: label for (label, idx) in sorted(d.items())}
            min_label, max_label, num_labels = min(labels), max(labels), len(labels)
            assert 0 == min_label
            if num_labels - 1 < max_label:
                print("Vocab error: not all indices used? max label: %i" % max_label, file=log.v1)
                print("unused labels: %r" % ([i for i in range(max_label + 1) if i not in labels],), file=log.v2)
            assert num_labels - 1 == max_label
            self.num_labels = len(labels)
            self._vocab = d
            self._labels = [label for (idx, label) in sorted(labels.items())]
            self._cache[filename] = (self._vocab, self._labels)

    @classmethod
    def create_vocab_dict_from_labels(cls, labels):
        """
        This is exactly the format which we expect when we read it in self._parse_vocab.

        :param list[str] labels:
        :rtype: dict[str,int]
        """
        d = {label: idx for (idx, label) in enumerate(labels)}
        assert len(d) == len(labels), "some labels are provided multiple times"
        return d

    @classmethod
    def create_vocab_from_labels(cls, labels, **kwargs):
        """
        Creates a `Vocabulary` from the given labels. Depending on whether the labels are identified as
        bytes, characters or words a `Utf8ByteTargets`, `CharacterTargets` or `Vocabulary` vocab is created.

        :param list[str] labels:
        :rtype: Vocabulary
        """
        kwargs = kwargs.copy()
        kwargs.setdefault("unknown_label", None)
        if len(labels) < 1000 and all([len(label) == 1 for label in labels]):
            # are these actually ordered raw bytes? -> assume utf8
            if all([ord(label) <= 255 and ord(label) == idx for idx, label in enumerate(labels)]):
                return Utf8ByteTargets()
            return CharacterTargets(vocab_file=None, labels=labels, **kwargs)
        return Vocabulary(vocab_file=None, labels=labels, **kwargs)

    def tf_get_init_variable_func(self, var):
        """
        :param tensorflow.Variable var:
        :rtype: (tensorflow.Session)->None
        """
        import tensorflow as tf
        from returnn.tf.util.basic import VariableAssigner

        assert isinstance(var, tf.Variable)
        assert var.dtype.base_dtype == tf.string
        assert var.shape.as_list() == [self.num_labels]
        assert len(self._labels) == self.num_labels

        def init_vocab_var(session):
            """
            :param tensorflow.Session session:
            """
            VariableAssigner(var).assign(session=session, value=self._labels)

        return init_vocab_var

    def to_id(
        self,
        label: Union[str, int, None],
        default: Union[str, Type[KeyError], None] = KeyError,
        allow_none: bool = False,
    ) -> Optional[int]:
        """
        :param label:
        :param default:
        :param allow_none: whether label can be None. in this case, None is returned
        """
        if isinstance(label, str):
            return self.label_to_id(label, default=default)
        if isinstance(label, int):
            if label == -1 and allow_none:
                return None
            if self.is_id_valid(label):
                return label
            if default is KeyError:
                raise KeyError("invalid label id %i" % label)
            return None
        if label is None and allow_none:
            return None
        raise TypeError("invalid label type %r" % type(label))

    def label_to_id(self, label: str, default: Union[int, Type[KeyError], None] = KeyError) -> Optional[int]:
        """
        :param label:
        :param default:
        """
        if default is KeyError:
            return self._vocab[label]
        return self._vocab.get(label, default)

    def id_to_label(self, idx: int, default: Union[str, Type[KeyError], None] = KeyError) -> Optional[str]:
        """
        :param idx:
        :param default:
        """
        if self.is_id_valid(idx):
            return self._labels[idx]
        if default is KeyError:
            raise KeyError("idx %i out of range" % idx)
        return default

    def is_id_valid(self, idx: int) -> bool:
        """
        :param idx:
        """
        return 0 <= idx < len(self._labels)

    @property
    def labels(self) -> List[str]:
        """list of labels"""
        return self._labels

    def get_seq(self, sentence: str) -> List[int]:
        """
        :param sentence: assumed to be seq of vocab entries separated by whitespace
        :return: seq of label indices
        """
        segments = sentence.split()
        return self.get_seq_indices(segments) + self.seq_postfix

    def get_seq_indices(self, seq: List[str]) -> List[int]:
        """
        :param seq: seq of labels (entries in vocab)
        :return: seq of label indices, returns unknown_label_id if unknown_label is set
        """
        if self.unknown_label is not None:
            return [self._vocab.get(k, self.unknown_label_id) for k in seq]
        return [self._vocab[k] for k in seq]

    def get_seq_labels(self, seq: Union[List[int], numpy.ndarray]) -> str:
        """
        Inverse of :func:`get_seq`.

        :param seq: 1D sequence of label indices
        :return: serialized sequence string, such that ``get_seq(get_seq_labels(seq)) == seq``
        """
        labels = self.labels
        return " ".join(map(labels.__getitem__, seq))


class BytePairEncoding(Vocabulary):
    """
    Vocab based on Byte-Pair-Encoding (BPE).
    This will encode the text on-the-fly with BPE.

    Reference:
    Rico Sennrich, Barry Haddow and Alexandra Birch (2016). Neural Machine Translation of Rare Words with Subword Units.
    Proceedings of the 54th Annual Meeting of the Association for Computational Linguistics (ACL 2016). Berlin, Germany.
    """

    def __init__(self, vocab_file, bpe_file, seq_postfix=None, **kwargs):
        """
        :param str vocab_file:
        :param str bpe_file:
        :param list[int]|None seq_postfix: labels will be added to the seq in self.get_seq
        """
        super(BytePairEncoding, self).__init__(vocab_file=vocab_file, seq_postfix=seq_postfix, **kwargs)
        from returnn.util.bpe import StandardBytePairEncoder

        self.bpe = StandardBytePairEncoder(bpe_codes_file=bpe_file, labels=self._labels)

    def get_seq(self, sentence):
        """
        :param str sentence:
        :rtype: list[int]
        """
        segments = self.bpe.segment_sentence(sentence)
        seq = self.get_seq_indices(segments)
        return seq + self.seq_postfix


class SamplingBytePairEncoding(Vocabulary):
    """
    Vocab based on Byte-Pair-Encoding (BPE).
    Like :class:`BytePairEncoding`, but here we randomly sample from different possible BPE splits.
    This will encode the text on-the-fly with BPE.
    """

    def __init__(
        self,
        vocab_file: str,
        breadth_prob: float,
        seq_postfix: Optional[List[int]] = None,
        label_postfix_merge_symbol: Optional[str] = NotSpecified,
        word_prefix_symbol: Optional[str] = NotSpecified,
        **kwargs,
    ):
        """
        :param vocab_file:
        :param breadth_prob:
        :param seq_postfix: labels will be added to the seq in self.get_seq
        :param label_postfix_merge_symbol: If given, will use this as label postfix merge symbol,
            i.e. when this occurs at the end of a label, it is supposed to be merged with the next label,
            i.e. the space between them is removed and is not a word boundary.
            If None, will not use any postfix merge symbol.
            If not specified, and also word_prefix_symbol is not specified, will use "@@" by default here,
            the standard from subword-nmt, and our original behavior.
        :param word_prefix_symbol: If given, every new word starts with this symbol.
            This also implies that there are no spaces between words
            and this symbol is a placeholder for the space.
            If None, will not use this logic.
            For SentencePiece, you usually would use "▁" here.
        """
        super(SamplingBytePairEncoding, self).__init__(vocab_file=vocab_file, seq_postfix=seq_postfix, **kwargs)
        from returnn.util.bpe import SamplingBytePairEncoder, BpePostMergeSymbol, BpeOpts

        if label_postfix_merge_symbol is NotSpecified and word_prefix_symbol is NotSpecified:
            label_postfix_merge_symbol = BpePostMergeSymbol
            word_prefix_symbol = None
        else:
            if label_postfix_merge_symbol is NotSpecified:
                label_postfix_merge_symbol = None
            if word_prefix_symbol is NotSpecified:
                word_prefix_symbol = None
        if word_prefix_symbol is not None:
            # I'm not sure if this makes sense otherwise...
            assert label_postfix_merge_symbol is None, (
                f"{self}: word_prefix_symbol {word_prefix_symbol},"
                f" label_postfix_merge_symbol {label_postfix_merge_symbol}"
            )

        self.rnd = numpy.random.RandomState(0)
        self.bpe = SamplingBytePairEncoder(
            labels=self._labels,
            breadth_prob=breadth_prob,
            rnd=self.rnd,
            unknown_label=self.id_to_label(self.unknown_label_id) if self.unknown_label_id is not None else None,
            opts=BpeOpts(label_postfix_merge_symbol=label_postfix_merge_symbol, word_prefix_symbol=word_prefix_symbol),
        )

    def set_random_seed(self, seed):
        """
        :param int seed:
        """
        self.rnd.seed(seed)

    def get_seq(self, sentence):
        """
        :param str sentence:
        :rtype: list[int]
        """
        segments = self.bpe.segment_sentence(sentence)
        seq = self.get_seq_indices(segments)
        return seq + self.seq_postfix


class SentencePieces(Vocabulary):
    """
    Uses the SentencePiece software,
    which supports different kind of subword units (including BPE, unigram, ...).

    https://github.com/google/sentencepiece/
    https://github.com/google/sentencepiece/tree/master/python

    Dependency::

      pip3 install --user sentencepiece

    """

    def __init__(self, **opts):
        """
        :param str model_file: The sentencepiece model file path.
        :param str model_proto: The sentencepiece model serialized proto.
        :param type out_type: output type. int or str. (Default = int)
        :param bool add_bos: Add <s> to the result (Default = false)
        :param bool add_eos: Add </s> to the result (Default = false)
          <s>/</s> is added after reversing (if enabled).
        :param bool reverse: Reverses the tokenized sequence (Default = false)
        :param bool enable_sampling: (Default = false)
        :param int nbest_size: sampling parameters for unigram. Invalid for BPE-Dropout.
          nbest_size = {0,1}: No sampling is performed.
          nbest_size > 1: samples from the nbest_size results.
          nbest_size < 0: (Default). assuming that nbest_size is infinite and samples
            from the all hypothesis (lattice) using
            forward-filtering-and-backward-sampling algorithm.
        :param float alpha: Soothing parameter for unigram sampling, and dropout probability of
          merge operations for BPE-dropout. (Default = 0.1)
        :param dict[str,str|int]|None control_symbols:
          https://github.com/google/sentencepiece/blob/master/doc/special_symbols.md
        :param dict[str,str|int]|None user_defined_symbols:
          https://github.com/google/sentencepiece/blob/master/doc/special_symbols.md
        """
        import sentencepiece as spm  # noqa

        self._opts = opts
        self._cache_key = opts.get("model_file", None)
        control_symbols = opts.pop("control_symbols", None)
        user_defined_symbols = opts.pop("user_defined_symbols", None)
        self.sp = spm.SentencePieceProcessor(**opts)  # noqa
        super(SentencePieces, self).__init__(
            vocab_file=None,
            seq_postfix=None,
            unknown_label=self.sp.unk_id(),
            eos_label=self.sp.eos_id(),
            bos_label=self.sp.bos_id(),
            pad_label=self.sp.pad_id(),
            control_symbols=control_symbols,
            user_defined_symbols=user_defined_symbols,
        )

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self._opts)

    def _parse_vocab(self):
        self.num_labels = self.sp.vocab_size()
        # Do not load labels/vocab here. This is not really needed.

    @property
    def labels(self) -> List[str]:
        """list of labels"""
        if self._cache_key and self._cache_key in self._cache:
            self._vocab, self._labels = self._cache[self._cache_key]
            assert self.num_labels == len(self._vocab) == len(self._labels)
        else:
            self._labels = [self.sp.id_to_piece(i) for i in range(self.num_labels)]  # noqa
            self._vocab = {label: i for (i, label) in enumerate(self._labels)}
            if self._cache_key:
                self._cache[self._cache_key] = (self._vocab, self._labels)
        return self._labels

    def is_id_valid(self, idx: int) -> bool:
        """
        :param idx:
        """
        return not self.sp.IsUnused(idx)

    def id_to_label(self, idx: int, default: Union[str, Type[KeyError], None] = KeyError) -> Optional[str]:
        """
        :param idx:
        :param default:
        """
        if default is not KeyError and not self.is_id_valid(idx):
            return default
        return self.sp.IdToPiece(idx)

    def label_to_id(self, label: str, default: Union[int, Type[KeyError], None] = KeyError) -> Optional[int]:
        """
        :param label:
        :param default:
        """
        res = self.sp.PieceToId(label)
        if res == self.unknown_label_id or res < 0 or res is None:
            # It could be that the label really is the unknown-label, or it could be that the label is unknown.
            if label == self.id_to_label(self.unknown_label_id):
                return self.unknown_label_id
            if default is KeyError:
                raise KeyError("label %r not found" % label)
            return default
        return res

    def set_random_seed(self, seed: int):
        """
        :param seed:
        """
        # Unfortunately, there is only a global seed,
        # and also, it will only be used for new threads
        # where the random generator was not used yet...
        # https://github.com/google/sentencepiece/issues/635
        import sentencepiece as spm  # noqa

        spm.set_random_generator_seed(seed)

    def get_seq(self, sentence: str) -> List[int]:
        """
        :param sentence: assumed to be seq of vocab entries separated by whitespace
        """
        return self.sp.encode(sentence, out_type=int)  # noqa


class CharacterTargets(Vocabulary):
    """
    Uses characters as target labels.
    Also see :class:`Utf8ByteTargets`.
    """

    def __init__(self, vocab_file, seq_postfix=None, unknown_label="@", labels=None, **kwargs):
        """
        :param str|None vocab_file:
        :param list[int]|None seq_postfix: labels will be added to the seq in self.get_seq
        :param str|None unknown_label:
        :param list[str]|None labels:
        """
        super(CharacterTargets, self).__init__(
            vocab_file=vocab_file, seq_postfix=seq_postfix, unknown_label=unknown_label, labels=labels, **kwargs
        )

    def get_seq(self, sentence):
        """
        :param str sentence:
        :rtype: list[int]
        """
        if self.unknown_label is not None:
            seq = [self._vocab.get(k, self.unknown_label_id) for k in sentence]
        else:
            seq = [self._vocab[k] for k in sentence]
        return seq + self.seq_postfix

    def get_seq_labels(self, seq):
        """
        :param list[int]|numpy.ndarray seq: 1D sequence
        :rtype: str
        """
        return "".join(map(self._labels.__getitem__, seq))


class Utf8ByteTargets(Vocabulary):
    """
    Uses bytes as target labels from UTF8 encoded text. All bytes (0-255) are allowed.
    Also see :class:`CharacterTargets`.
    """

    def __init__(self, seq_postfix=None, **opts):
        """
        :param list[int]|None seq_postfix: labels will be added to the seq in self.get_seq
        """
        super(Utf8ByteTargets, self).__init__(vocab_file=None, seq_postfix=seq_postfix, unknown_label=None, **opts)

    def _parse_vocab(self):
        """
        Sets self.vocab, self.labels, self.num_labels.
        """
        self._vocab = {chr(i): i for i in range(256)}
        self._labels = [chr(i) for i in range(256)]
        self.num_labels = 256

    def get_seq(self, sentence):
        """
        :param str sentence:
        :rtype: list[int]
        """
        if sys.version_info[0] >= 3:
            seq = list(sentence.encode("utf8"))
        else:
            seq = list(bytearray(sentence.encode("utf8")))
        return seq + self.seq_postfix

    def get_seq_labels(self, seq):
        """
        :param list[int]|numpy.ndarray seq: 1D sequence
        :rtype: str
        """
        if not isinstance(seq, numpy.ndarray):
            seq = numpy.array(seq)
        assert seq.ndim == 1
        if seq.dtype != numpy.uint8:
            assert ((seq >= 0) & (seq < 256)).all(), f"invalid byte value, must be within 0-255: {seq}"
            seq = seq.astype(numpy.uint8)
        return bytearray(seq).decode(encoding="utf8")



"""
Basic utilities for RETURNN, such as some Numpy utilities, or system utilities, or other helpers.
Independent from other parts of RETURNN.
Also independent from TensorFlow or Theano (see :mod:`returnn.tf.util` or :mod:`returnn.theano.util`).
"""

# Some basic imports.
from .basic import BackendEngine, NumbersDict, BehaviorVersion
